import time

from flasgger import SwaggerView

from onto import retrofit
import pytest
from .fixtures import CTX


@pytest.mark.skip
def test_retrofit(CTX):

    class ExampleView(SwaggerView):
        def get(self):
            return {"0": {
                "name": "Hello"
            }}

    retrofit._retrofit(view=ExampleView(), collection_path="hellos")

    # time.sleep(5)



import asyncio
from collections import defaultdict
from typing import AsyncGenerator


class Broker:
    """
    Broker

    WARNING: unsubscribe is not implemented. RAM size will instead increase when the
    subscriber exited, since the stored topics are not consumed.

    RECOMMENDATION: (work-around) limit RAM size so that the container would reboot when out of memory.

    """

    def __init__(self):
        self.source_queue = asyncio.Queue()
        self.source_sink_mapping = defaultdict(list)
        self.sink_queue = defaultdict(asyncio.Queue)

    def bind(self, *, pub_id, sub_id):
        self.source_sink_mapping[pub_id].append(sub_id)

    async def publish(self, topic_name, item):
        await self.source_queue.put((topic_name, item))

    async def source_while_loop(self):
        while True:
            pub_id, item = await self.source_queue.get()
            for sub_id in self.source_sink_mapping[pub_id]:
                q: asyncio.Queue = self.sink_queue[sub_id]
                await q.put(item)

    async def sink_while_loop(self, sub_id) -> AsyncGenerator:
        while True:
            self.sink_queue[sub_id]: asyncio.Queue
            item = self.sink_queue[sub_id].get()
            yield item
            self.sink_queue[sub_id].task_done()



from onto.context import Context as CTX
from .base import Sink


class FirestoreSink(Sink):

    def emit(self, reference, snapshot):
        CTX.db.set(ref=reference, snapshot=snapshot)


class ViewModelSink(Sink):

    def emit(self, obj):
        CTX.db.set(ref=obj.doc_ref, snapshot=obj.to_snapshot())


class FormSink(ViewModelSink):

    def emit(self, view_model_obj):
        view_model_obj.propagate_change()


class DomainModelSink(FirestoreSink):

    def emit(self, obj, **kwargs):
        obj.save(**kwargs)





class Sink:

    def emit(self):
        pass

    def close(self):
        pass

    def emit_and_close(self, *args, **kwargs):
        pass

    def insert(self):
        pass

    def update(self):
        pass

    def upsert(self):
        pass

    def remove(self):
        pass



"""
DomainModel source:
    Invokes a mediator with instantiated domain model
"""
from typing import Type

from onto.database import Reference
from onto.domain_model import DomainModel
from onto.source.firestore import FirestoreSource
from onto.context import Context as CTX


class DomainModelSource(FirestoreSource):
    """
    TODO: limit maxsize
    """

    def __init__(self, domain_model_cls: Type[DomainModel], *args, **kwargs):
        self.domain_model_cls = domain_model_cls
        query = self.domain_model_cls.get_query().where(*args, **kwargs)
        super().__init__(query=query)

    def _call(self, container):
        with container.lock:
            for func_name, ref, snapshot in self.delta(container):
                obj = self.domain_model_cls.from_snapshot(
                    ref=ref, snapshot=snapshot)
                self._invoke_mediator(func_name=func_name, obj=obj)


class DomainModelTransactionalSource(DomainModelSource):

    def _call(self, container):
        with container.lock:
            for func_name, ref, snapshot in self.delta(container):
                self._operation(func_name=func_name, ref=ref, snapshot=snapshot)

    def _operation(self, func_name, ref, snapshot):
        _transaction = CTX.db.firestore_client.transaction()

        from google.cloud.firestore_v1 import transactional
        @transactional
        def _op(transaction):
            obj = self.domain_model_cls.get(
                doc_ref=ref, transaction=transaction)
            self._invoke_mediator(
                func_name=func_name,
                obj=obj,
                transaction=transaction
            )
        _op(_transaction)


class DomainModelPathSource(DomainModelSource):

    def _call(self, ref: Reference):
        doc_id = ref.last
        doc_ref = self.domain_model_cls.ref_from_id(doc_id=doc_id)
        snapshot = CTX.db.get(ref=doc_ref)
        return self.domain_model_cls.from_snapshot(
            ref=doc_ref, snapshot=snapshot)




from onto.context import Context as CTX


from functools import wraps


def run_transaction(func):
    """ Decorator to run a function in transaction. All instances
            of subclasses of FirestoreObject will default to this transaction
            for reading and writing to Firestore.

    :return:
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        # Set transaction context variable
        transaction = CTX.db.transaction()
        token = CTX.transaction_var.set(transaction)
        from google.cloud import firestore
        @firestore.transactional
        def new_func(transaction: firestore.Transaction):
            return func(*args, **kwargs, transaction=transaction)
        res = new_func(transaction)
        # Reverse transaction context variable
        CTX.transaction_var.reset(token)
        return res
    return wrapper




from functools import lru_cache
from typing import Iterable, Tuple

from onto.attrs.attribute_new import AttributeBase


def _make_schema_name(cls):
    return f"_{cls.__name__}_GeneratedSchema"


def _schema_cls_from_attributed_class(cls):
    """ Make schema from a class containing AttributeBase+ objects

    :return:
    """

    import_only = getattr(cls.Meta, "import_only", False)
    export_only = getattr(cls.Meta, "export_only", False)

    d = dict()
    for key, attr in _collect_attrs(cls):
        if key == '_pk_':
            """ ponyorm duplicates primary key field 
            TODO: make better 
            """
            continue

        try:

            from onto.attrs.unit import MonadContext
            context = MonadContext.context()

            # context = context.marshmallow_capable_base()

            case_conversion = True
            if hasattr(cls.Meta, "case_conversion"):
                new_case_conversion = cls.Meta.case_conversion
                if new_case_conversion is not None:
                    case_conversion = new_case_conversion

            context = context.attribute_name(key).data_key(key)

            annotation = attr.annotation
            context = context.annotate(annotation)

            from onto.utils import camel
            transformation = camel if case_conversion else lambda a: a
            context = context.data_key_from_name(transformation=transformation)

            if getattr(cls.Meta, 'is_sorter', False):
                context = context.string
                # context = context.data_key(key)

            with context:
                field = attr.marshmallow_field
        except Exception as e:
            raise ValueError(f'Error while making field {key}') from e
        field.load_only = field.load_only or import_only
        field.dump_only = field.dump_only or export_only
        d[key] = field
    if len(d) == 0:
        return None
    else:
        # TODO: note that inheritance of super.Meta replaces Meta completely
        # TODO:     as opposed to cls.Meta = super.Meta + cls.Meta
        if unwrap := getattr(cls.Meta, "unwrap", None):

            from onto.mapper.fields import Field
            field: Field = d[unwrap]
            data_key = field.data_key

            from marshmallow.decorators import pre_load, post_dump
            @pre_load
            def unwrap_field(self, data, many, **kwargs):
                if many:
                    raise ValueError  # TODO: check logics first
                new_data = {
                    data_key: data
                }
                return new_data

            d['unwrap_field'] = unwrap_field

            @post_dump
            def wrap_field(self, data, many, **kwargs):
                if many:
                    raise ValueError  # TODO: check logics first
                return data[data_key]  # TODO: check for key error for read only fields
            d['wrap_field'] = wrap_field

        m_meta = {
            "exclude": getattr(cls.Meta, "exclude", tuple()),
        }

        d["Meta"] = type(
            "Meta",
            tuple(),
            m_meta
        )

        # TODO: make better
        # if hasattr(cls.Meta, "case_conversion"):
        #     d["case_conversion"] = cls.Meta.case_conversion

        schema_base = cls._schema_base if cls._schema_cls is None else cls._schema_cls

        TempSchema = type(_make_schema_name(cls), (schema_base,), d)

        return TempSchema


def _graphql_type_from_py(t: type, input=False):
    import graphql
    PY_TYPE_MAP_GQL = {
        int: graphql.GraphQLInt,
        bool: graphql.GraphQLBoolean,
        float: graphql.GraphQLFloat,
        str: graphql.GraphQLString,
        list: graphql.GraphQLList
    }
    if t in PY_TYPE_MAP_GQL:
        return PY_TYPE_MAP_GQL[t]
    else:
        if input:
            return t.get_graphql_object_type(is_input=input)
        else:
            raise ValueError


def _graphql_field_from_attr(attr, input=False):
    import graphql

    if not input:
        field_base = graphql.GraphQLField
    else:
        field_base = graphql.GraphQLInputField

    if hasattr(attr, 'collection') and attr.collection == list:
        f = lambda ot: graphql.GraphQLList(ot)
    else:
        f = lambda ot: ot

    from onto import attrs
    if attr.__class__ is attrs.attribute.EmbeddedAttribute:
        e_cls = attr.type_cls
        e_graphql = e_cls.get_graphql_object_type(is_input=input)
        field = field_base(
            type_=f(e_graphql),
            description=attr.doc
        )
        return attr.data_key, field
    elif isinstance(attr, attrs.attribute.AttributeBase):
        import graphql
        field = field_base(
            type_=f(_graphql_type_from_py(t=attr.type_cls)),
            description=attr.doc
        )
        return attr.data_key, field
    else:
        raise NotImplementedError


ots = dict()
from collections import defaultdict
interfaces_of = defaultdict(list)
implementations_of = defaultdict(list)


# def _graphql_object_type_from_attributed_class(*args, **kwargs):
#     res = _graphql_object_type_from_attributed_class(*args, **kwargs)
#     ots[(args, kwargs)] = res
#     return res


def _get_graphql_ots():
    return ots


@lru_cache(maxsize=None)
def _graphql_object_type_from_attributed_class(cls, input=False, **kwargs):
    """ Make GraphQL schema from a class containing AttributeBase+ objects

    :return:
    """

    # import_only = getattr(cls.Meta, "import_only", False)
    # export_only = getattr(cls.Meta, "export_only", False)

    import graphql

    from onto.attrs.unit import MonadContext

    if not input:
        base = graphql.GraphQLObjectType
    else:
        base = graphql.GraphQLInputObjectType

    def convert(key, attr):
        if not attr.properties.is_internal:
            from onto.attrs.unit import MonadContext
            context = MonadContext.context()
            context = context.graphql_capable(is_input=input)
            context = context.attribute_name(key)

            annotation = attr.annotation
            context = context.annotate(annotation)

            context = context.data_key_from_name()
            context = context.optional

            with context:
                return attr.properties.data_key, attr.graphql_field

    def fields_gen():
        for key, attr in _collect_attrs(cls):
            # TODO: check to see if .attribute_name(key).data_key(key) should be removed
            if ret := convert(key=key, attr=attr):
                yield ret
        else:
            yield from ()

    if unwrap := getattr(cls.Meta, "unwrap", None):

        key, attr = next((k, v) for k, v in _collect_attrs(cls) if k == unwrap)
        data_key, graphql_field = convert(key=key, attr=attr)

        from graphql import GraphQLScalarType
        assert isinstance(graphql_field.type, GraphQLScalarType)

        this_type = GraphQLScalarType(
            cls.__name__,
            serialize=graphql_field.type.serialize,
            parse_value=graphql_field.type.parse_value,
            parse_literal=graphql_field.type.parse_literal,
        )
        return this_type
    else:
        fields = lambda: dict(fields_gen())
        type_name = cls.__name__

        if getattr(cls, '_union', None):
            for klass in cls._get_subclasses():
                interfaces_of[klass].append(cls)
            subclasses = {
                    klass: _graphql_object_type_from_attributed_class(klass, input=input, interfaces=lambda: [ot])
                    for klass in cls._get_subclasses()
                    if klass != cls
            }

            def resolve_type(__self, value, _type):
                return subclasses[__self.__class__]  # TODO: make better

            ot = graphql.GraphQLInterfaceType(
                name=type_name,
                resolve_type=resolve_type,
                fields=fields,
                **kwargs
            )

        else:

            # TODO: maybe add
            # if input:
            #     type_name += "Input"

            ot = base(
                type_name,
                fields=fields,
                **kwargs
            )

        ots[ot.name] = ot
        return ot


def _collect_attrs(cls) -> Iterable[Tuple[str, AttributeBase]]:
    """
    Collect all AttributeBase+ objects in the class and its ancestors.
    :param cls:
    :return:
    """
    import inspect
    from functools import partial

    for key, attr in cls.__attributes.items():
        # if issubclass(getattr(cls, key)):
        #     attr = getattr(cls, key)
        yield key, attr  # attr.bind_class_(_bound_cls=cls, _attr_name=key)
    else:
        yield from ()



from onto.source.protocol import Protocol
from unittest.mock import patch, call, ANY
from .fixtures import CTX

from examples.meeting_room.tests.fixtures import *


def test_register():

    protocol = Protocol()

    @protocol._register('on_update')
    @protocol._register('on_create')
    def foo():
        print('foo invoked')

    assert protocol.mapping['on_create'] == 'foo'
    assert protocol.mapping['on_update'] == 'foo'


def test_source(meeting, CTX):
    from onto.source.leancloud import BeforeSaveDomainModelSource as dms

    with patch.object(dms, '_invoke_mediator') as mock_method:
        import leancloud
        cla = leancloud.Object.extend('Meeting')
        cla_obj = cla(
            **meeting.to_dict()
        )

        from examples.meeting_room.domain_models import Meeting
        source = dms(Meeting)
        source._call(
            cla_obj
        )
        assert mock_method.call_args_list == [call(func_name='before_save', obj=ANY)]



import time

from google.cloud.firestore_v1 import DocumentSnapshot, DocumentReference

"""
_Watch is ported from from google.cloud.firestore_v1.watch

Copy of the license of google.cloud.firestore_v1.watch: 
# Copyright 2017 Google LLC All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""


# Copyright 2017 Google LLC All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import collections
import threading
import datetime
from enum import Enum
import functools

import pytz

from google.api_core.bidi import ResumableBidiRpc
from google.api_core.bidi import BackgroundConsumer
from google.cloud.firestore_v1.proto import firestore_pb2
from google.cloud.firestore_v1 import _helpers

from google.api_core import exceptions

import grpc

"""Python client for Google Cloud Firestore Watch."""

_LOGGER = logging.getLogger(__name__)

INITIAL_TARGET_ID = 0x0  # "Py"

GRPC_STATUS_CODE = {
    "OK": 0,
    "CANCELLED": 1,
    "UNKNOWN": 2,
    "INVALID_ARGUMENT": 3,
    "DEADLINE_EXCEEDED": 4,
    "NOT_FOUND": 5,
    "ALREADY_EXISTS": 6,
    "PERMISSION_DENIED": 7,
    "UNAUTHENTICATED": 16,
    "RESOURCE_EXHAUSTED": 8,
    "FAILED_PRECONDITION": 9,
    "ABORTED": 10,
    "OUT_OF_RANGE": 11,
    "UNIMPLEMENTED": 12,
    "INTERNAL": 13,
    "UNAVAILABLE": 14,
    "DATA_LOSS": 15,
    "DO_NOT_USE": -1,
}
_RPC_ERROR_THREAD_NAME = "Thread-OnRpcTerminated"
_RECOVERABLE_STREAM_EXCEPTIONS = (
    exceptions.Aborted,
    exceptions.Cancelled,
    exceptions.Unknown,
    exceptions.DeadlineExceeded,
    exceptions.ResourceExhausted,
    exceptions.InternalServerError,
    exceptions.ServiceUnavailable,
    exceptions.Unauthenticated,
)
_TERMINATING_STREAM_EXCEPTIONS = (exceptions.Cancelled,)

DocTreeEntry = collections.namedtuple("DocTreeEntry", ["value", "index"])


class WatchDocTree(object):
    # TODO: this class has been modified to suit the purpose of flask-boiler
    # TODO: note that the RAM usage will keep increasing since consumed
    #           document snapshot in _removed are not deleted
    #           so that we can find the original version of the document
    #           when a document is changed.

    def __init__(self):
        self._dict = {}
        self._index = 0
        self._removed = {}

    def keys(self):
        return [*self._removed.keys(), *self._dict.keys()]

    def _copy(self):
        wdt = WatchDocTree()
        wdt._dict = self._dict.copy()
        wdt._index = self._index
        wdt._removed = self._removed.copy()
        self = wdt
        return self

    def insert(self, key, value):
        self = self._copy()
        self._dict[key] = DocTreeEntry(value, self._index)
        self._index += 1
        return self

    def find(self, key):
        return self._dict[key]

    def remove(self, key):
        self = self._copy()
        self._removed[key] = self._dict[key]
        del self._dict[key]
        return self

    def __iter__(self):
        for k in self._dict:
            yield k

    def __len__(self):
        return len(self._dict)

    def __contains__(self, k):
        return k in self._dict


class ChangeType(Enum):
    ADDED = 1
    REMOVED = 2
    MODIFIED = 3


class DocumentChange(object):
    def __init__(self, type, document, old_index, new_index):
        """DocumentChange

        Args:
            type (ChangeType):
            document (document.DocumentSnapshot):
            old_index (int):
            new_index (int):
        """
        # TODO: spec indicated an isEqual param also
        self.type = type
        self.document = document
        self.old_index = old_index
        self.new_index = new_index


class WatchResult(object):
    def __init__(self, snapshot, name, change_type):
        self.snapshot = snapshot
        self.name = name
        self.change_type = change_type


def _maybe_wrap_exception(exception):
    """Wraps a gRPC exception class, if needed."""
    if isinstance(exception, grpc.RpcError):
        return exceptions.from_grpc_error(exception)
    return exception


def document_watch_comparator(doc1, doc2):
    assert doc1 == doc2, "Document watches only support one document."
    return 0


def _should_recover(exception):
    wrapped = _maybe_wrap_exception(exception)
    return isinstance(wrapped, _RECOVERABLE_STREAM_EXCEPTIONS)


def _should_terminate(exception):
    wrapped = _maybe_wrap_exception(exception)
    return isinstance(wrapped, _TERMINATING_STREAM_EXCEPTIONS)


class _Watch(object):

    BackgroundConsumer = BackgroundConsumer  # FBO unit tests
    ResumableBidiRpc = ResumableBidiRpc  # FBO unit tests

    def __init__(
        self,
        firestore,
        # target,
        comparator,
        # snapshot_callback,
        document_snapshot_cls,
        document_reference_cls,
        BackgroundConsumer=None,  # FBO unit testing
        ResumableBidiRpc=None,  # FBO unit testing
    ):
        """
        Args:
            firestore:
            target: can be None
            comparator:
            snapshot_callback: Callback method to process snapshots.
                Args:
                    docs (List(DocumentSnapshot)): A callback that returns the
                        ordered list of documents stored in this snapshot.
                    changes (List(str)): A callback that returns the list of
                        changed documents since the last snapshot delivered for
                        this watch.
                    read_time (string): The ISO 8601 time at which this
                        snapshot was obtained.

            document_snapshot_cls: instance of DocumentSnapshot
            document_reference_cls: instance of DocumentReference
        """
        # self._document_reference = document_reference
        self._firestore = firestore
        self._api = firestore._firestore_api
        self._targets = dict()
        self._target_callbacks = dict()
        # if target is not None:
        #     target_id = target['target_id']
        #     self._targets[target_id] = target
        self._comparator = comparator
        self.DocumentSnapshot = document_snapshot_cls
        self.DocumentReference = document_reference_cls
        # self._snapshot_callback = snapshot_callback
        self._closing = threading.Lock()
        self._closed = False

        self.resume_token = None

        # rpc_request = self._get_rpc_request

        if ResumableBidiRpc is None:
            ResumableBidiRpc = self.ResumableBidiRpc  # FBO unit tests

        self._rpc = ResumableBidiRpc(
            self._api.transport.listen,
            should_recover=_should_recover,
            should_terminate=_should_terminate,
            # initial_request=rpc_request,
            metadata=self._firestore._rpc_metadata,
        )

        # TODO: recover this line somewhere
        # if self.resume_token is not None:
        # # self._targets["resume_token"] = self.resume_token

        self._rpc.add_done_callback(self._on_rpc_done)

        # Initialize state for on_snapshot
        # The sorted tree of QueryDocumentSnapshots as sent in the last
        # snapshot. We only look at the keys.
        self.doc_tree = WatchDocTree()

        # A map of document names to QueryDocumentSnapshots for the last sent
        # snapshot.
        self.doc_map = {}

        # The accumulates map of document changes (keyed by document name) for
        # the current snapshot.
        self.change_map = {}

        from collections import defaultdict
        self.change_log = defaultdict(list)

        # The current state of the query results.
        self.current = False

        # We need this to track whether we've pushed an initial set of changes,
        # since we should push those even when there are no changes, if there
        # aren't docs.
        self.has_pushed = False

        # The server assigns and updates the resume token.
        if BackgroundConsumer is None:  # FBO unit tests
            BackgroundConsumer = self.BackgroundConsumer

        self._consumer = BackgroundConsumer(self._rpc, self.on_snapshot)
        self._consumer.start()

    # def _get_rpc_request(self):
    #     if self.resume_token is not None:
    #         self._targets["resume_token"] = self.resume_token
    #     if INITIAL_TARGET_ID in self._targets:
    #         return firestore_pb2.ListenRequest(
    #             database=self._firestore._database_string,
    #             add_target=self._targets
    #         )
    #     else:
    #         return firestore_pb2.ListenRequest(
    #             database=self._firestore._database_string
    #         )

    def add_target(self, target, callback):
        """ Only new targets can be added.

        :param target:
        :return:
        """
        target_id = target['target_id']
        if target_id in self._targets:
            raise ValueError
        self._targets[target_id] = target
        self._target_callbacks[target_id] = callback
        self._rpc.send(
            firestore_pb2.ListenRequest(
                database=self._firestore._database_string,
                add_target=self._targets[target_id]
            )
        )

    @property
    def is_active(self):
        """bool: True if this manager is actively streaming.

        Note that ``False`` does not indicate this is complete shut down,
        just that it stopped getting new messages.
        """
        return self._consumer is not None and self._consumer.is_active

    def close(self, reason=None):
        """Stop consuming messages and shutdown all helper threads.

        This method is idempotent. Additional calls will have no effect.

        Args:
            reason (Any): The reason to close this. If None, this is considered
                an "intentional" shutdown.
        """
        with self._closing:
            if self._closed:
                return

            # Stop consuming messages.
            if self.is_active:
                _LOGGER.debug("Stopping consumer.")
                self._consumer.stop()
            self._consumer = None

            self._rpc.close()
            self._rpc = None
            self._closed = True
            _LOGGER.debug("Finished stopping manager.")

        if reason:
            # Raise an exception if a reason is provided
            _LOGGER.debug("reason for closing: %s" % reason)
            if isinstance(reason, Exception):
                raise reason
            raise RuntimeError(reason)

    def _on_rpc_done(self, future):
        """Triggered whenever the underlying RPC terminates without recovery.

        This is typically triggered from one of two threads: the background
        consumer thread (when calling ``recv()`` produces a non-recoverable
        error) or the grpc management thread (when cancelling the RPC).

        This method is *non-blocking*. It will start another thread to deal
        with shutting everything down. This is to prevent blocking in the
        background consumer and preventing it from being ``joined()``.
        """
        _LOGGER.info("RPC termination has signaled manager shutdown.")
        future = _maybe_wrap_exception(future)
        thread = threading.Thread(
            name=_RPC_ERROR_THREAD_NAME, target=self.close, kwargs={"reason": future}
        )
        thread.daemon = True
        thread.start()

    def unsubscribe(self):
        self.close()

    @classmethod
    def for_document(
        cls,
        document_ref,
        snapshot_callback,
        snapshot_class_instance,
        reference_class_instance,
    ):
        """
        Creates a watch snapshot listener for a document. snapshot_callback
        receives a DocumentChange object, but may also start to get
        targetChange and such soon

        Args:
            document_ref: Reference to Document
            snapshot_callback: callback to be called on snapshot
            snapshot_class_instance: instance of DocumentSnapshot to make
                snapshots with to pass to snapshot_callback
            reference_class_instance: instance of DocumentReference to make
                references

        """
        return cls(
            document_ref,
            document_ref._client,
            {
                "documents": {"documents": [document_ref._document_path]},
                "target_id": INITIAL_TARGET_ID,
            },
            document_watch_comparator,
            snapshot_callback,
            snapshot_class_instance,
            reference_class_instance,
        )

    @classmethod
    def for_query(
        cls, query, snapshot_callback, snapshot_class_instance, reference_class_instance
    ):
        parent_path, _ = query._parent._parent_info()
        query_target = firestore_pb2.Target.QueryTarget(
            parent=parent_path, structured_query=query._to_protobuf()
        )

        return cls(
            query,
            query._client,
            {"query": query_target, "target_id": INITIAL_TARGET_ID},
            query._comparator,
            snapshot_callback,
            snapshot_class_instance,
            reference_class_instance,
        )

    def _on_snapshot_target_change_no_change(self, proto):
        _LOGGER.debug("on_snapshot: target change: NO_CHANGE")
        change = proto.target_change

        no_target_ids = change.target_ids is None or len(change.target_ids) == 0

        """
        If empty, the change applies to all targets.
        """

        # TODO: move resume token key in self._targets to elsewhere
        target_ids = list(self._targets.keys()) if no_target_ids \
            else change.target_ids

        if no_target_ids and change.read_time and self.current:
            # TargetChange.CURRENT followed by TargetChange.NO_CHANGE
            # signals a consistent state. Invoke the onSnapshot
            # callback as specified by the user.
            self.push(change.read_time, change.resume_token)

    def _on_snapshot_target_change_add(self, proto):
        pass
        # _LOGGER.debug("on_snapshot: target change: ADD")
        # target_id = proto.target_change.target_ids[0]
        # if target_id not in self._targets:
        #     raise RuntimeError("Unexpected target ID %s sent by server" % target_id)
        # if target_id != WATCH_TARGET_ID:
        #     raise RuntimeError("Unexpected target ID %s sent by server" % target_id)

    def _on_snapshot_target_change_remove(self, proto):
        pass
        # _LOGGER.debug("on_snapshot: target change: REMOVE")
        # change = proto.target_change
        #
        # code = 13
        # message = "internal error"
        # if change.cause:
        #     code = change.cause.code
        #     message = change.cause.message
        #
        # message = "Error %s:  %s" % (code, message)
        #
        # raise RuntimeError(message)

    def _on_snapshot_target_change_reset(self, proto):
        # Whatever changes have happened so far no longer matter.
        _LOGGER.debug("on_snapshot: target change: RESET")
        self._reset_docs()

    def _on_snapshot_target_change_current(self, proto):
        _LOGGER.debug("on_snapshot: target change: CURRENT")
        self.current = True

    def on_snapshot(self, proto):
        """
        Called everytime there is a response from listen. Collect changes
        and 'push' the changes in a batch to the customer when we receive
        'current' from the listen response.

        Args:
            listen_response(`google.cloud.firestore_v1.types.ListenResponse`):
                Callback method that receives a object to
        """
        _LOGGER.info(proto)
        TargetChange = firestore_pb2.TargetChange

        target_changetype_dispatch = {
            TargetChange.NO_CHANGE: self._on_snapshot_target_change_no_change,
            TargetChange.ADD: self._on_snapshot_target_change_add,
            TargetChange.REMOVE: self._on_snapshot_target_change_remove,
            TargetChange.RESET: self._on_snapshot_target_change_reset,
            TargetChange.CURRENT: self._on_snapshot_target_change_current,
        }

        target_change = getattr(proto, "target_change", "")
        document_change = getattr(proto, "document_change", "")
        document_delete = getattr(proto, "document_delete", "")
        document_remove = getattr(proto, "document_remove", "")
        filter_ = getattr(proto, "filter", "")

        if str(target_change):
            target_change_type = target_change.target_change_type
            _LOGGER.debug("on_snapshot: target change: " + str(target_change_type))
            meth = target_changetype_dispatch.get(target_change_type)
            if meth is None:
                _LOGGER.info(
                    "on_snapshot: Unknown target change " + str(target_change_type)
                )
                self.close(
                    reason="Unknown target change type: %s " % str(target_change_type)
                )
            else:
                try:
                    meth(proto)
                except Exception as exc2:
                    _LOGGER.debug("meth(proto) exc: " + str(exc2))
                    raise

            # NOTE:
            # in other implementations, such as node, the backoff is reset here
            # in this version bidi rpc is just used and will control this.

        elif str(document_change):
            _LOGGER.debug("on_snapshot: document change")

            target_ids = document_change.target_ids or []

            for target_id in target_ids:
                self.change_log[target_id].append(proto)

        # NB: document_delete and document_remove (as far as we, the client,
        # are concerned) are functionally equivalent

        elif str(document_delete):
            _LOGGER.debug("on_snapshot: document change: DELETE")
            name = document_delete.document
            self.change_map[name] = ChangeType.REMOVED

            target_ids = document_delete.removed_target_ids or []

            for target_id in target_ids:
                self.change_log[target_id].append(proto)

        elif str(document_remove):
            _LOGGER.debug("on_snapshot: document change: REMOVE")
            name = document_remove.document
            self.change_map[name] = ChangeType.REMOVED

            target_ids = document_change.target_ids or []

            for target_id in target_ids:
                self.change_log[target_id].append(proto)

        elif filter_:
            _LOGGER.debug("on_snapshot: filter update")
            if filter_.count != self._current_size():
                # We need to remove all the current results.
                self._reset_docs()
                # The filter didn't match, so re-issue the query.
                # TODO: reset stream method?
                # self._reset_stream();

        elif proto is None:
            self.close()
        else:
            _LOGGER.debug("UNKNOWN TYPE. UHOH")
            self.close(reason=ValueError("Unknown listen response type: %s" % proto))

    def push(self, read_time, next_resume_token):
        """
        Assembles a new snapshot from the current set of changes and invokes
        the user's callback. Clears the current changes on completion.
        """
        # deletes, adds, updates = _Watch._extract_changes(
        #     self.doc_map, self.change_map, read_time
        # )
        #
        # updated_tree, updated_map, appliedChanges = self._compute_snapshot(
        #     self.doc_tree, self.doc_map, deletes, adds, updates
        # )

        # TODO: It is possible in the future we will have the tree order
        # on insert. For now, we sort here.
        # key = functools.cmp_to_key(self._comparator)
        # keys = sorted(updated_tree.keys(), key=key)
        for target_id, changes in self.change_log.items():
            if target_id not in self._target_callbacks:
                continue
            callback = self._target_callbacks[target_id]
            callback(target_id, changes, read_time)
        self.has_pushed = True
        #
        # self.doc_tree = updated_tree
        # self.doc_map = updated_map
        self.change_log.clear()
        self.change_map.clear()
        self.resume_token = next_resume_token

    @staticmethod
    def _extract_changes(doc_map, changes, read_time):
        deletes = []
        adds = []
        updates = []

        for name, value in changes.items():
            if value == ChangeType.REMOVED:
                if name in doc_map:
                    deletes.append(name)
            elif name in doc_map:
                if read_time is not None:
                    value.read_time = read_time
                updates.append(value)
            else:
                if read_time is not None:
                    value.read_time = read_time
                adds.append(value)

        return (deletes, adds, updates)

    def _compute_snapshot(
        self, doc_tree, doc_map, delete_changes, add_changes, update_changes
    ):
        updated_tree = doc_tree
        updated_map = doc_map

        assert len(doc_tree) == len(doc_map), (
            "The document tree and document map should have the same "
            + "number of entries."
        )

        def delete_doc(name, updated_tree, updated_map):
            """
            Applies a document delete to the document tree and document map.
            Returns the corresponding DocumentChange event.
            """
            assert name in updated_map, "Document to delete does not exist"
            old_document = updated_map.get(name)
            # TODO: If a document doesn't exist this raises IndexError. Handle?
            existing = updated_tree.find(old_document)
            old_index = existing.index
            updated_tree = updated_tree.remove(old_document)
            del updated_map[name]
            return (
                DocumentChange(ChangeType.REMOVED, old_document, old_index, -1),
                updated_tree,
                updated_map,
            )

        def add_doc(new_document, updated_tree, updated_map):
            """
            Applies a document add to the document tree and the document map.
            Returns the corresponding DocumentChange event.
            """
            name = new_document.reference._document_path
            assert name not in updated_map, "Document to add already exists"
            updated_tree = updated_tree.insert(new_document, None)
            new_index = updated_tree.find(new_document).index
            updated_map[name] = new_document
            return (
                DocumentChange(ChangeType.ADDED, new_document, -1, new_index),
                updated_tree,
                updated_map,
            )

        def modify_doc(new_document, updated_tree, updated_map):
            """
            Applies a document modification to the document tree and the
            document map.
            Returns the DocumentChange event for successful modifications.
            """
            name = new_document.reference._document_path
            assert name in updated_map, "Document to modify does not exist"
            old_document = updated_map.get(name)
            if old_document.update_time != new_document.update_time:
                remove_change, updated_tree, updated_map = delete_doc(
                    name, updated_tree, updated_map
                )
                add_change, updated_tree, updated_map = add_doc(
                    new_document, updated_tree, updated_map
                )
                return (
                    DocumentChange(
                        ChangeType.MODIFIED,
                        new_document,
                        remove_change.old_index,
                        add_change.new_index,
                    ),
                    updated_tree,
                    updated_map,
                )

            return None, updated_tree, updated_map

        # Process the sorted changes in the order that is expected by our
        # clients (removals, additions, and then modifications). We also need
        # to sort the individual changes to assure that old_index/new_index
        # keep incrementing.
        appliedChanges = []

        key = functools.cmp_to_key(self._comparator)

        # Deletes are sorted based on the order of the existing document.
        delete_changes = sorted(delete_changes)
        for name in delete_changes:
            change, updated_tree, updated_map = delete_doc(
                name, updated_tree, updated_map
            )
            appliedChanges.append(change)

        add_changes = sorted(add_changes, key=key)
        _LOGGER.debug("walk over add_changes")
        for snapshot in add_changes:
            _LOGGER.debug("in add_changes")
            change, updated_tree, updated_map = add_doc(
                snapshot, updated_tree, updated_map
            )
            appliedChanges.append(change)

        update_changes = sorted(update_changes, key=key)
        for snapshot in update_changes:
            change, updated_tree, updated_map = modify_doc(
                snapshot, updated_tree, updated_map
            )
            if change is not None:
                appliedChanges.append(change)

        assert len(updated_tree) == len(updated_map), (
            "The update document "
            + "tree and document map should have the same number of entries."
        )
        return (updated_tree, updated_map, appliedChanges)

    def _affects_target(self, target_ids, current_id):
        if target_ids is None:
            return True

        return current_id in target_ids

    def _current_size(self):
        """
        Returns the current count of all documents, including the changes from
        the current changeMap.
        """
        deletes, adds, _ = _Watch._extract_changes(self.doc_map, self.change_map, None)
        return len(self.doc_map) + len(adds) - len(deletes)

    def _reset_docs(self):
        """
        Helper to clear the docs on RESET or filter mismatch.
        """
        _LOGGER.debug("resetting documents")
        self.change_map.clear()
        self.resume_token = None

        # Mark each document as deleted. If documents are not deleted
        # they will be sent again by the server.
        for snapshot in self.doc_tree.keys():
            name = snapshot.reference._document_path
            self.change_map[name] = ChangeType.REMOVED

        self.current = False


# class _Watch(Watch):
#     pass
#     # def _on_snapshot_target_change_remove(self, proto):
#     #     """
#     #     target removed: assuming once=True and on_snapshot invoked
#     #     :param proto:
#     #     :return:
#     #     """
#     #     pass

class DataListener:

    def __init__(self, document_refs=None, snapshot_callback=None, firestore=None, once=False, query=None):
        """

        :param document_refs: a list of string
        :param snapshot_callback:
        """

        def comparator(doc1: DocumentSnapshot, doc2: DocumentSnapshot):
            """ TODO: use carefully

            :param doc1:
            :param doc2:
            :return:
            """
            return 1
            # if doc1.reference._document_path > doc2.reference._document_path:
            #     return 1
            # elif doc1.reference._document_path == doc2.reference._document_path:
            #     return 0
            # else:
            #     return -1

        if document_refs is not None:

            target = {
                    "documents": {"documents": document_refs},
                    "target_id": WATCH_TARGET_ID,
                    "once": once
                }

            self.watch = _Watch(document_reference=None,
                                firestore=firestore,
                                target=target,
                                comparator=comparator,
                                snapshot_callback=snapshot_callback,
                                document_snapshot_cls=DocumentSnapshot,
                                document_reference_cls=DocumentReference,
                                )

        elif query is not None:

            parent_path, _ = query._parent._parent_info()
            query_target = firestore_pb2.Target.QueryTarget(
                parent=parent_path, structured_query=query._to_protobuf()
            )
            target = {"query": query_target, "target_id": WATCH_TARGET_ID}

            self.watch = _Watch(document_reference=query,
                                firestore=query._client,
                                target=target,
                                comparator=comparator,
                                snapshot_callback=snapshot_callback,
                                document_snapshot_cls=DocumentSnapshot,
                                document_reference_cls=DocumentReference,
                                )
        else:
            raise ValueError

    def wait_for_once_done(self):
        """
        TODO: Note that for now this method only detects >= 1 updates, instead
            of exactly one update.
        :return:
        """
        # TODO: Find a better way
        # TODO: review and test
        # time.sleep(2)
        while not self.watch.current:
            # Note: the smallest interval that time.sleep can
            #   be invoked with differs by OS
            time.sleep(.020)  # Check back in 20ms
        # self.watch._consumer._thread.join()

    def __del__(self):
        self.watch.close()



from onto.mapper.fields import StructuralRef
from onto.mapper.schema import SchemaBase
from onto.firestore_object import FirestoreObjectValMixin
from onto.models.base import Serializable
from onto.store.struct import struct_ref

"""
Store provides a unified interface to access dependencies from a View Model
Shelf provides a key-value store with reference as key and instance of 
    Domain Model as value (single moment in time) 
Snapshot Container stores snapshots and manages versions of the snapshots 
    (all moment in time) 

The same reference and snapshot may map to different obj_type. 
Thus shelf layer keeps obj_type and reference as key, and 
    objects as values. 
"""


class BPSchema(SchemaBase):

    @classmethod
    def from_dict(cls, fields, name):
        for _, val in fields.items():
            if not issubclass(val.__class__, StructuralRef):
                raise TypeError
        return super().from_dict(fields=fields, name=name)

    @property
    def structural_ref_fields(self):
        return [fd for _, fd in self.fields.items() if isinstance(fd, StructuralRef)]


class Store(FirestoreObjectValMixin, Serializable):
    class Meta:
        case_conversion = False

    _schema_base = BPSchema

    @classmethod
    def from_struct(cls, struct):
        from onto.store import Gallery

        schema_obj = cls.get_schema_obj()
        d = schema_obj.load(struct)
        _store = Gallery()
        d = cls._import_from_dict(d, _store=_store, transaction=None)
        _store.refresh(transaction=None)
        instance = cls.new(**d)  # TODO: fix unexpected arguments
        return instance

    @classmethod
    def from_objects(cls, **kwargs):
        d = {
            key: struct_ref(obj=val)
            for key, val in kwargs.items()
        }
        return cls.from_struct(d)

    @classmethod
    def from_snapshot_struct(cls, snapshot_struct, **kwargs):
        struct, container = snapshot_struct.to_struct_and_container()
        store = cls(**kwargs)
        store.struct = struct
        store._container = container
        return store

    # def add_snapshot(self, key, dm_cls, snapshot: DocumentSnapshot):
    #     struct_sub, container = self.struct, self._container
    #     if "." in "key":
    #         parent_name, key = key.split(".")
    #         struct_sub = struct_sub[parent_name]
    #     struct_sub[key] = (dm_cls, snapshot.reference.id)
    #     container.set(snapshot.reference._document_path, snapshot)

    def __init__(
            self, *args, readonly=False, transaction=None, _store=None,
            **kwargs):
        self.transaction = transaction
        self.readonly = readonly
        super().__init__(*args, **kwargs)

    # def __init__(self, obj_options=None, **kwargs):
    #     """
    #
    #     :param struct:
    #     :param snapshot_container:
    #     :param obj_options: keyword arguments to pass to snapshot_to_obj
    #         (eventually applied to obj_cls.from_dict)
    #     """
    #     super().__init__(**kwargs)
    #     if obj_options is None:
    #         obj_options = dict()
    #     self._obj_options = obj_options
    #     self._gallery = Gallery()
    #     self.struct = Struct(schema_obj=self.schema_obj)
    #     # self._info = self._get_manifests(self.struct, self.schema_obj)
    #     self.objs = dict()

    def propagate_back(self):
        """
        TODO: add transaction
        :return:
        """
        self._export_as_dict(_store=self._store, transaction=None)
        self._store.save()



import pytest
from flasgger import Swagger
from flask import Flask


@pytest.fixture
def CTX():
    """
    Note that pytest.fixture(scope="package") is experimental according
        to pytest documentations
    :return:
    """

    from onto.context import Context as TST_CTX
    from onto.config import Config

    if TST_CTX.config is None:
        config = Config.load()
        TST_CTX.read(config)

    return TST_CTX

@pytest.fixture
def setup_app(CTX):

    app = Flask(__name__)
    swagger = Swagger(app)

    return app



from collections import defaultdict

from onto.attrs.attribute import RelationshipAttribute
from onto.common import _NA
from typing import Tuple

from onto.mapper.fields import StructuralRef
from onto.mapper.schema import SchemaBase
from onto.utils import snapshot_to_obj
from onto.context import Context as CTX


def to_ref(dm_cls, dm_doc_id):
    """
    TODO: check doc_ref._document_path alternatives that are compatible
        with firestore listeners

    :param val:
    :return:
    """
    from onto.database.firestore import FirestoreReference

    doc_ref: FirestoreReference = dm_cls._get_collection() / dm_doc_id

    return CTX.db.make_document_path(doc_ref)


class BusinessPropertyStore:

    @classmethod
    def from_snapshot_struct(cls, snapshot_struct, **kwargs):
        struct, container = snapshot_struct.to_struct_and_container()
        store = BusinessPropertyStore(struct=struct,
                                      snapshot_container=container,
                                      **kwargs
                                      )
        store.refresh()
        return store

    def __init__(self, struct, snapshot_container, obj_options=None):
        """

        :param struct:
        :param snapshot_container:
        :param obj_options: keyword arguments to pass to snapshot_to_obj
            (eventually applied to obj_cls.from_dict)
        """
        super().__init__()
        if obj_options is None:
            obj_options = dict()
        self._obj_options = obj_options
        self._container = snapshot_container
        self.struct = struct
        self.schema_obj = struct.schema_obj
        self._g, self._gr, self._manifest = \
            self._get_manifests(self.struct, self.schema_obj)
        self.objs = dict()

    @property
    def bprefs(self):
        return {doc_ref for (_, doc_ref) in self._manifest.copy()}

    def refresh(self):
        for obj_cls, doc_ref in self._manifest:
            self.objs[doc_ref] = snapshot_to_obj(
                self._container.get(doc_ref),
                super_cls=obj_cls,
                **self._obj_options
            )

    def __getattr__(self, item):

        if item not in self._g:
            raise AttributeError

        if isinstance(self._g[item], dict):
            return {
                k: self.objs[v]
                for k, v in self._g[item].items()
            }
        else:
            return self.objs[self._g[item]]

    def propagate_back(self):
        for _, obj in self.objs.items():
            obj.save()

    @staticmethod
    def _get_manifests(struct, schema_obj) -> Tuple:

        g, gr, manifest = dict(), defaultdict(list), set()

        for fd in schema_obj.structural_ref_fields:
            key = fd.attribute
            val = struct.get_item(key)

            if val is None:
                continue

            if fd.many:
                g[key] = dict()
                for k, v in val.items():
                    if "." in k:
                        raise ValueError
                    doc_ref = to_ref(*v)
                    g[key][k] = doc_ref
                    gr[doc_ref].append("{}.{}".format(key, k))
                    manifest.add( (v[0], doc_ref) )
            else:
                doc_ref = to_ref(*val)
                g[key] = doc_ref
                gr[doc_ref].append(key)
                manifest.add( (val[0], doc_ref) )

        return dict(g), dict(gr), manifest




import onto.models.base


class RModelSup(onto.models.base.BaseRegisteredModel):
    pass


class RModelA(RModelSup):
    pass


class RModelAB(RModelA):
    pass


class RModelB(RModelSup):
    pass


def test_get_cls_from_name():
    assert RModelSup.get_cls_from_name("RModelSup") == RModelSup


def test_tree():
    assert RModelSup._tree["RModelSup"] == {"RModelA", "RModelB"}


def test_get_children():

    assert RModelSup._get_children() == {RModelA, RModelB}
    assert RModelA._get_parents() == {RModelSup,}


def test_get_subclasses():
    assert RModelSup._get_subclasses() == {RModelSup, RModelA, RModelAB, RModelB}


def test_get_subclasses_str():
    assert RModelSup._get_subclasses_str() == \
           ["RModelA", "RModelAB", "RModelB", "RModelSup"]


def test_register_subclass():

    class CModelParent(onto.models.base.BaseRegisteredModel):
        pass

    class CModelChild(CModelParent):
        pass

    assert CModelChild.__name__ == "CModelChild"



"""
Author: Zixuan Rao
Reference: https://github.com/Amsterdam/subsidieservice/blob/master/subsidy_service/subsidy_service/context.py

Usage:
1. In testing or production, initialize Context singleton first by calling config.Context.read() in __init__.py of
    caller package
2. In package (test_data_access as an example) __init__.py, add "import config"
3. In files under the package: CTX = test_data_access.config.Context

"""

import logging
import os

# from google.cloud import storage

# This import is intended to be accessed with google.cloud.logging.*


import logging

from contextvars import ContextVar
from onto.database import Database, Listener

_transaction_var: ContextVar['firestore.Transaction'] = \
    ContextVar('_transaction_var', default=None)


class Context:
    """ Context Singleton for Firestore, Firebase and Celery app.
    TODO: consider development behavior inconsistency for different
            values of debug and testing. (Behaviors of context
            may be different in production as compared to that in
            testing)

        Example:

    NOTE: reload should not be called for more than once
    TODO: enforce call reload only once and rename methods starting with reload

    """

    # from onto.config import Config

    debug = None
    firebase_app: 'firebase_admin.App' = None
    db: Database = None
    config = None
    # celery_app: Celery = None
    logger: logging.Logger = None
    listener: Listener = None

    transaction_var = _transaction_var

    # Deleted on purpose to ensure that .<flag> is not evaluated as False
    #   when .<flag> is neither set to True, nor set to False.
    # debug = None
    # testing = None
    # _cred = None
    __instance = None

    _ready = False
    """
    _ready is True when Context is loaded (and/or is read) 
    """

    def __init__(self, *args, **kwargs):
        raise NotImplementedError('Do not initialize this class, use the class methods and properties instead. ')

    @classmethod
    def _reload_logger(cls):
        """ Loads logger.
        TODO: call Logger.setLevel(...)  with level specified by user
        """
        cls.logger = logging.getLogger()
        cls.logger.setLevel(level=logging.ERROR)

    # @staticmethod
    # def _logger_type(logger):
    #     if isinstance(logger, google.cloud.logging.logger.Logger):
    #         return 'gcloud-logger'
    #     elif isinstance(logger, logging.Logger):
    #         return 'sys-logger'
    #     else:
    #         raise TypeError

    @classmethod
    def _enable_logging(cls, output_level=logging.DEBUG, output_choice=None,
                        certificate_path=None):
        """ Set cls.logger to output to standard out.
        NOTE: cls._reload_logger should be called first.
        level applies when output_choice is sys-logger, and does not
            apply to cls.logger

        :param level:
        :return:
        """
        if output_choice == 'sys-logger':
            import sys
            root = cls.logger

            handler = logging.StreamHandler(sys.stdout)
            handler.setLevel(output_level)
            formatter = logging.Formatter(
                '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            handler.setFormatter(formatter)
            root.addHandler(handler)
        elif output_choice == 'gcloud-logger':
            import google.cloud.logging
            logging_client = google.cloud.logging.Client.from_service_account_json(
                json_credentials_path=certificate_path
            )
            # "Retrieves a Cloud Logging handler based on the environment
            # you're running in and integrates the handler with the
            # Python logging module. By default this captures all logs
            # at INFO level and higher"
            # ref: gcloud logging docs
            logging_client.get_default_handler()
            logging_client.setup_logging()
        else:
            raise TypeError

    @classmethod
    def load(cls) -> None:
        from onto.config import Config
        config = Config.load()
        cls.read(config)

    @classmethod
    def read(cls, config) -> None:
        """ Description
            Read config file andd reload firebase app and firestore client.

        NOTE: ordering of cls._reload_* methods is important

        :type cls:
        :param cls:

        :raises:

        :rtype:
        """
        # if cls.config == config:
        #     """
        #     Skip reading config if the current config is the same as
        #         the config to read.
        #     """
        #     return cls
        if cls._ready:
            """ Only allow one read/load 
            """
            import warnings
            warnings.warn('Config is read more than once ')
        cls.config = config

        # """
        # TODO: NOTE that PROXY is used here
        # Best be replaced for Safety reasons
        # ATTENTION:
        # """
        # os.environ["HTTP_PROXY"] = "https://34.85.42.121:8899"
        # os.environ["HTTPS_PROXY"] = "https://34.85.42.121:8899"

        """
        Configure logger to output to sys-logger during test 
        and gcloud-logger during staging and etc 
        """
        cls._reload_logger()
        if cls.config.TESTING:
            cls._enable_logging(output_level=logging.DEBUG,
                                output_choice='sys-logger')
        else:
            cls._enable_logging(output_choice='gcloud-logger',
                                certificate_path=cls.config.FIREBASE_CERTIFICATE_JSON_PATH)

        cls._reload_debug_flag(cls.config.DEBUG)
        cls._reload_testing_flag(cls.config.TESTING)
        cls._reload_dbs(cls.config.database)
        default_db_name = cls.config.default_database
        cls._reload_default_db(db_name=default_db_name)
        cls._reload_services(cls.config.services)  # leancloud.engine depends on db

        cls.logger.info(f"onto.Context has finished "
                        f"loading config: {vars(config)}")
        cls._ready = True

    @classmethod
    def _reload_services(cls, services: dict):
        class Services:
            pass

        cls.services = Services()
        for svc_name, svc_config in services.items():
            svc = cls.create_service(svc_config)
            setattr(cls.services, svc_name, svc)

    @classmethod
    def _reload_dbs(cls, database: dict):
        class Dbs:
            pass

        cls.dbs = Dbs()
        for db_name, db_config in database.items():
            db = cls.create_db(db_config)
            setattr(cls.dbs, db_name, db)

    @classmethod
    def _reload_default_db(cls, db_name=None):
        if db_name is not None:
            db = getattr(cls.dbs, db_name)
            setattr(cls, 'db', db)

    @staticmethod
    def create_db(db_config):
        print(db_config)
        if db_config['type'] == "couch":
            server = db_config['server']
            try:
                import couchdb
            except ImportError as e:
                raise TypeError('couchdb server is configured, but '
                                'importing couchdb module has failed') from e
            return couchdb.Server(server)
        elif db_config['type'] == 'couchbase':
            try:
                from couchbase.cluster import Cluster, ClusterOptions
                from couchbase_core.cluster import PasswordAuthenticator
            except ImportError as e:
                raise TypeError('couchbase server is configured, but '
                                'importing couchbase module has failed') from e

            cluster = Cluster(db_config['cluster'], ClusterOptions(
                PasswordAuthenticator(
                    db_config['username'], db_config['password']
                )
            ))
            bucket = cluster.bucket(db_config['bucket'])
            return bucket
        elif db_config['type'] == 'firestore':
            caller_module_path = os.path.curdir
            cert_path = os.path.join(
                caller_module_path, db_config['certificate_path']
            )
            # import requests
            # px = {
            #     'http': "35.189.144.254:3128",
            #     'https': "35.189.144.254:3128"
            # }
            # session = requests.Session()
            # session.proxies = px
            def _ua():
                import pkg_resources
                version = pkg_resources.get_distribution("onto").version
                return f'onto/{version}'
            def _client_info():
                import google.api_core.gapic_v1.client_info
                import pkg_resources
                _GAPIC_LIBRARY_VERSION = pkg_resources.get_distribution(
                    "google-cloud-firestore"
                ).version
                return google.api_core.gapic_v1.client_info.ClientInfo(
                    gapic_version=_GAPIC_LIBRARY_VERSION,
                    user_agent=_ua()
                )

            from google.cloud import firestore
            client = firestore.Client.from_service_account_json(
                cert_path, client_info=_client_info())
            from onto.database.firestore import FirestoreDatabase
            FirestoreDatabase.firestore_client = client
            return FirestoreDatabase
        elif db_config['type'] == 'leancloud':
            try:
                import leancloud
            except ImportError as e:
                raise TypeError('leancloud server is configured, but '
                                'importing leancloud module has failed') from e
            if 'app_id' in db_config and 'master_key' in db_config:
                """
                Allow empty init for Leancloud engine uses 
                """
                leancloud.init(
                    app_id=db_config['app_id'],
                    master_key=db_config['master_key']
                )
            else:
                import leancloud

                APP_ID = os.environ['LEANCLOUD_APP_ID']
                APP_KEY = os.environ['LEANCLOUD_APP_KEY']
                MASTER_KEY = os.environ['LEANCLOUD_APP_MASTER_KEY']

                leancloud.init(APP_ID, app_key=APP_KEY, master_key=MASTER_KEY)
                # 如果需要使用 master key 权限访问 LeanCLoud 服务，请将这里设置为 True
                leancloud.use_master_key(False)

            from onto.database.leancloud import LeancloudDatabase
            return LeancloudDatabase
        elif db_config['type'] == 'kafka':
            try:
                bootstrap_servers = db_config['bootstrap_servers']
                from onto.database.kafka import KafkaDatabase
                KafkaDatabase.bootstrap_servers = bootstrap_servers
                return KafkaDatabase
            except ImportError as e:
                raise TypeError('kafka is configured, but '
                                'importing kafka module has failed') from e
        elif db_config['type'] == 'pony':
            try:
                from pony.orm import Database
                db = Database()
                from pony.orm.dbproviders.sqlite import SQLiteProvider
                db.bind(SQLiteProvider, filename=':memory:', create_db=True)
                return db
            except ImportError as e:
                raise TypeError('pony is configured, but '
                                'importing pony module has failed') from e
        elif db_config['type'] == 'mock':
            from onto.database.mock import MockDatabase
            return MockDatabase
        else:
            raise ValueError

    @classmethod
    def create_service(cls, svc_config):
        service_type = svc_config['type']
        if service_type == 'engine':
            try:
                import leancloud
            except ImportError as e:
                raise TypeError('leancloud server is configured, but '
                                'importing leancloud module has failed') from e
            engine = leancloud.Engine()
            return engine
        elif service_type == 'firebase':
            try:
                import firebase_admin
                app_name = svc_config['app_name']
                storage_bucket_name = svc_config.get('storage_bucket_name', None)
                if storage_bucket_name is None:
                    storage_bucket_name = f"{app_name}.appspot.com"

                cred = cls._get_credentials(svc_config['certificate_filename'])
                return firebase_admin.initialize_app(
                    credential=cred,
                    name=app_name,
                    options={
                        'storageBucket': storage_bucket_name
                    }
                )
            except Exception as e:
                logging.exception('Error initializing firebase_app')
                raise e
        elif service_type == 'celery':
            from celery import Celery
            celery_app = Celery('tasks',
                                    broker='pyamqp://guest@localhost//')
            return celery_app
        elif service_type == 'authing':
            user_pool_id = svc_config['user_pool_id']
            secret = svc_config['secret']  # TODO: isolate side effect of assignment
            from authing import Authing
            authing = Authing({
                "userPoolId": user_pool_id,
                "secret": secret,
            })
            return authing
        else:
            raise ValueError

    @classmethod
    def _reload_debug_flag(cls, debug):
        cls.debug = debug

    @classmethod
    def _reload_testing_flag(cls, testing):
        """
        When testing is set to True, all authenticate decorators in services returns uid as "testuid1"
        :param testing:
        :return:
        """
        cls.testing = testing

    @classmethod
    def _get_credentials(cls, certificate_path):
        try:
            from firebase_admin import credentials
            _cred = credentials.Certificate(certificate_path)
            return _cred
        except Exception as e:
            logging.exception('Error initializing credentials.Certificate')
            raise e
        # TODO delete certificate path in function call



import functools
from queue import Queue
from threading import Thread


class Coordinator:
    """ Manages a task queue.
    Listener adds function invocations to the task queue.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.q = Queue()
        self._start_thread()

    # def start(self):
    #     self._start_thread()

    def _main(self):
        """ Push None to self.q to stop the thread

        :return:
        """
        while True:
            item = self.q.get()
            if item is None:
                break
            try:
                item()
            except Exception as e:
                from onto.context import Context as CTX
                CTX.logger.exception(f"a task in the queue has failed {item}")
            self.q.task_done()

    def _start_thread(self):
        self.thread = Thread(target=self._main, daemon=True)
        self.thread.start()

    def _add_awaitable(self, item):
        """ TODO: test

        :param item:
        :return:
        """
        self.q.put(item)



from onto.database import Listener, Reference, Snapshot


class GenericListener(Listener):
    from asyncio.queues import Queue
    from collections import defaultdict
    #
    # def create_queue():
    #     from onto.context import Context as CTX

    qs = defaultdict(Queue)

    @classmethod
    def _pub(cls, reference: Reference, snapshot: Snapshot):
        col = reference.collection
        cls.qs[col].put_nowait((reference, snapshot))

    @classmethod
    async def _sub(cls, col):
        while True:
            item = await cls.qs[col].get()
            if item is None:
                break
            try:
                yield item
            except Exception as e:
                from onto.context import Context as CTX
                CTX.logger.exception(f"a task in the queue has failed {item}")
            # cls.qs[col].task_done()

    @classmethod
    async def listen(cls, col, source):
        async for ref, snapshot in cls._sub(col):
            if snapshot is not None:
                await source._invoke_mediator(
                    func_name='on_create', ref=ref, snapshot=snapshot)
            else:
                # delete
                await source._invoke_mediator(
                    func_name='on_delete', ref=ref, snapshot=snapshot)



import pytest

# from tests.utils import _delete_all

def pytest_sessionstart(session):
    from onto.context import Context as CTX
    from onto.config import Config

    if CTX.config is None:
        config = Config.load()
        CTX.read(config)
    #
    # _delete_all(CTX, "Orbit")
    # _delete_all(CTX, "locations")
    # _delete_all(CTX, "orbits")
    # _delete_all(CTX, "rideHosts")
    # _delete_all(CTX, "riderBookings")
    # _delete_all(CTX, "riderTargets")
    # _delete_all(CTX, "users")
    #
    # _delete_all(CTX, "RainbowDAV")
    # _delete_all(CTX, "Shard")
    # _delete_all(CTX, "UserViewDAV")
    # _delete_all(CTX, "categories")
    # _delete_all(CTX, "counters")
    # _delete_all(CTX, "hellos")
    #
    # _delete_all(CTX, subcollection_name="MeetingSessionDAV_PATCH")
    # _delete_all(CTX, subcollection_name="_PATCH_UserViewDAV")
    # _delete_all(CTX, subcollection_name="MeetingSessionDAV")


def pytest_sessionfinish(session, exitstatus):
    from onto.context import Context as CTX
    #
    # if exitstatus == 0:
    #     _delete_all(CTX, "Orbit")
    #     _delete_all(CTX, "locations")
    #     _delete_all(CTX, "orbits")
    #     _delete_all(CTX, "rideHosts")
    #     _delete_all(CTX, "riderBookings")
    #     _delete_all(CTX, "riderTargets")
    #     _delete_all(CTX, "users")
    #
    #     _delete_all(CTX, "RainbowDAV")
    #     _delete_all(CTX, "Shard")
    #     _delete_all(CTX, "UserViewDAV")
    #     _delete_all(CTX, "categories")
    #     _delete_all(CTX, "counters")
    #     _delete_all(CTX, "hellos")
    #
    #     _delete_all(CTX, subcollection_name="MeetingSessionDAV_PATCH")
    #     _delete_all(CTX, subcollection_name="_PATCH_UserViewDAV")
    #     _delete_all(CTX, subcollection_name="MeetingSessionDAV")





class ConnectorBase:
    """
    Connector invokes instance methods of a mediator

    Order:
    1. External component invokes connector._invoke_mediator
    2. _invoke_mediator(*args, **kwargs)
        i.  Locate the instance method of mediator to call: f
        ii. Transform with side effect (*args, **kwargs) -> (*nargs, **nkwargs)
        iii.Call f(*nargs, **nkwargs) -> ret
        iv. Transform with side effect ret -> nret
        v.  return nret

    connector.

    """

    # router: None

    def __init__(self):
        from pymonad.reader import Compose
        self._before_call = Compose(lambda x: x)
        self._after_call = Compose(lambda x: x)

    def before_call(self, f):
        self._before_call = self._before_call.map(f)
        return self

    def after_call(self, f):
        self._after_call = self._after_call.map(f)
        return self

    def invoke_mediator(self, *args, **kwargs):
        from pymonad.reader import Compose
        applicative = Compose(lambda x: x)
        applicative = applicative.map(self._before_call)
        applicative = applicative.map(self._invoke_mediator)
        applicative = applicative.map(self._after_call)
        return applicative(*args, **kwargs)

    def _invoke_mediator(self, *args, **kwargs):
        pass



"""
We want a singleton instance that routes all requests and triggers including
    WebSocket, NoSQL change stream, cloud function triggers,
    HTTP request, and RPC.

App instance will keep a event loop, and places handlers on the event loop
    when triggered. We also want it to support job graph, as we may need
    to retract or cancel earlier invocations when new data has come in and
    we need to invoke with the new data in its place.

"""


from onto.source.rest import Source, RestProtocol


class StarletteConnector(Source):
    """
    @some_connector.triggers.route('/callback', ('POST',))
    async def some_method(self, request):
    """

    _protocol_cls = RestProtocol

    def route(self, *args, **kwargs):
        return self.protocol.route(*args, **kwargs)
        # protocol: RestProtocol = self.protocol

    def __init__(self, *args, **kwargs):
        self._mediator_instance = None
        super().__init__(*args, **kwargs)

    # def emit(self, res):
    #     abort(res)

    @property
    def mediator_instance(self):
        if self._mediator_instance is None:
            self._mediator_instance = self.parent()()
        return self._mediator_instance

    def start(self, url_prefix=None):
        from starlette.routing import Route

        routes = list()
        # bp = Blueprint(self.parent().__name__, __name__, url_prefix=url_prefix)
        for rule, methods, f_name in self.protocol._get_functions():
            async def routable(*args, **kwargs):
                return await self._invoke_mediator_async(func_name=f_name, *args,  **kwargs)

            route = Route(path=rule, endpoint=routable, methods=methods)
            routes.append(route)

        from starlette.routing import Mount
        mount = Mount(url_prefix, routes=routes)

        return [mount]


from flask import Flask

from .fixtures import CTX, setup_app
from .color_fixtures import color_refs, rainbow_vm
from onto.view import rest_api


def test_rainbow_stuffs(CTX, setup_app, color_refs, rainbow_vm):
    app = setup_app

    assert isinstance(app, Flask)

    test_client = app.test_client()

    class RainbowViewModel(rainbow_vm):
       pass

    mediator = rest_api.ViewMediator(
        view_model_cls=RainbowViewModel,
        app=app
    )
    mediator.add_instance_get(rule="/rainbow/<string:color_names>")

    # time.sleep(1)

    res = test_client.get(
        path="/rainbow/yellow+magenta+cian")

    assert res.json == {
        'rainbowName': 'cian-magenta-yellow',
        'colors': ['cian', 'magenta', 'yellow']
    }



import typing
from collections import namedtuple
from functools import partial, lru_cache

try:
    from pony.orm.core import Attribute, PrimaryKey, Discriminator, Optional, Required
except ImportError:
    import warnings
    warnings.warn("pony import skipped")

from onto.mapper import fields
from typing import Type, Callable

from onto.common import _NA
from onto.query.cmp import Condition, RootCondition

from functools import cached_property

_ATTRIBUTE_STORE_NAME = "_attrs"

class ValueNotProvided:
    pass


class TypeClsAsArgMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class MakePonyAttribute:

    def _make_pony_attribute_cls(self):
        if self.__class__.__name__ == "Discriminator":
            # TODO: make better
            return Discriminator
        if self.import_required:
            return Required
        else:
            return Optional

    # @lru_cache(maxsize=None)
    def _make_pony_attribute(self):
        column_name = self.data_key
        py_type = self.type_cls
        is_required = self.import_required
        _pony_attribute_cls = self._make_pony_attribute_cls()
        _pony_attribute = _pony_attribute_cls(
            py_type, is_required=is_required, column=column_name)
        return _pony_attribute

    @cached_property
    def _pony_attribute(self):
        if self.is_concrete:
            return self._make_pony_attribute()
        else:
            return None

    @property
    def is_collection(self):
        return self._pony_attribute.is_collection


udf = namedtuple('udf', ['f', 'fid', 'd', 'result_pytype'])


class MakePonyFunction:

    assigned_ids = set()

    @classmethod
    def _random_func_id(cls, N=5):
        import random, string
        fid = '_onto_udf_' + ''.join(random.choices(string.ascii_lowercase, k=N))
        if fid in cls.assigned_ids:
            raise ValueError("Collision")
        cls.assigned_ids.add(fid)
        return fid

    @classmethod
    def register_function(cls, f, type_cls):
        class Some:
            pass

        # TODO: add notes about `typing.*` not supported

        # import ast
        # code = f.__code__
        import inspect
        d = list(inspect.getclosurevars(f).unbound)

        def k(*args):
            _self = Some()
            for idx in range(len(d)):
                k = d[idx]
                v = args[idx]
                setattr(_self, k, v)
            _f = f
            from types import MethodType
            _f = MethodType(f, _self)
            return _f()

        fid = cls._random_func_id()
        udf_obj = udf(f=k, fid=fid, d=d, result_pytype=type_cls)
        # f_reg.append(udf_obj)
        return udf_obj

    @staticmethod
    def _flink_type_mapping(type_cls):
        from pyflink.table.types import _type_mappings
        return _type_mappings[type_cls]

    @classmethod
    def _register_flink_function(cls, udf_obj: udf):
        from pyflink.table import ScalarFunction, DataTypes
        from pyflink.table import udf as flink_udf
        F = type(udf_obj.fid, bases=(ScalarFunction,), dict={
            'eval': udf_obj.f,
        })
        table_env.create_temporary_function(
            udf_obj.fid,
            flink_udf(
                F(),
                input_types=[cls._flink_type_mapping(v) for _, v in udf_obj.d.items()],  # TODO: ensure order of keys
                result_type=cls._flink_type_mapping(udf_obj.result_pytype)
            )
        )
        # flink_udf()

#
# @db.on_connect(provider='flink')
# def _add_udf(db, connection):
#     # cursor = connection.cursor()
#
#     def create_function(name, num_params, func):
#         from pony.orm.dbproviders.sqlite import keep_exception
#         func = keep_exception(func)
#         connection.create_function(name, num_params, func)
#
#     for udf in f_reg:
#         create_function(udf.fid, len(udf.d), udf.f)


class AttributeBase(RootCondition):

    def __hash__(self):
        # TODO: debug
        return id(self)

    def _make_field_cls(self):
        # Makes marshmallow field cls
        return fields.Field

    def _make_field(self) -> fields.Field:
        """
        TODO: implement
        :return:
        """
        field_cls = self._make_field_cls()
        return field_cls(**self._field_kwargs, attribute=self.name)

    def _get_data_key(self):
        """
        TODO: deprecate
        :return:
        """
        return self.data_key

    def __set_name__(self, owner, name):
        self.parent = owner
        self.name = name

    def copy(self):
        """
        TODO: debug
        NOTE: contents in self.field_kwargs may or may not be copied
        TODO: fix
        :return:
        """
        from copy import deepcopy
        return deepcopy(self)

    def _get_select(self, instance):
        return getattr(instance, self.name)

    def __get__(self, instance, owner):
        """ Only allow attribute object to be invoked "get" on
                a class, and not an instance.

        :param instance:
        :param owner:
        :return:
        """
        if instance is None:
            return self
        else:
            raise AttributeError()

    @property
    def data_key(self):
        if self._data_key is not None:
            return self._data_key
        else:
            return self.name

    @property
    def is_internal(self):
        return not self.import_enabled and not self.export_enabled

    @property
    def pytype(self):
        return self.type_cls

    def __init__(
            self,
            *,
            initialize:bool= _NA,
            initializer: typing.Union[Callable[[object], None], _NA] = _NA,
            data_key=_NA,

            import_enabled:bool= _NA,
            import_default=_NA,
            import_required=_NA,

            export_enabled:bool= _NA,
            export_default=_NA,
            export_required=_NA,

            requires=_NA,

            type_cls=_NA,
            doc=_NA,
            is_concrete=_NA,
            validate=_NA,
            **kwargs

    ):
        """

        :param initialize: If true, initialize the value as the first step.
            The value may be set again later in the process of calling "new".
        :param initialize_value: The value to initialize the attribute to.
            May be a callable to avoid mutable default arguments.

        :param data_key: Sets import_from and export_to (field name in a
            document in the database)

        :param import_enabled: If true, the value will be imported
            to the object; Default to True.
        :param import_default: Import this value if the field
            name is missing from a document in the database
        :param import_required:

        :param export_enabled: If true, the value will be exported to
            a field in the database. Default to True.
        :param export_default: Export this value if attribute
            is missing from the object

        :param requires: A list of attributes that is required to be
            imported before this attribute is imported (do not pass
            in values that may result in a cycle, or you risk infinite
            call loop)  TODO: implement circular dependency detection

        :param type_cls: type for the attribute (no use for now)
        :param validate: validate arg to pass to marshmallow
        """
        super().__init__(**kwargs)
        field_kwargs = dict()

        field_kwargs["allow_none"] = True

        """
        Initialization precedes import 
        """
        if initialize is _NA:
            initialize = False

        self.initialize = initialize

        """
        Initializer: used to initialize the attribute when 
            attr.initialize is set to True. 
        """
        if initializer is _NA:
            def _initializer_not_defined(_self):
                raise NotImplementedError
            initializer = _initializer_not_defined

        self.initializer = initializer

        # Parse data key
        if data_key != _NA:
            self._data_key = data_key
            field_kwargs["data_key"] = data_key
        else:
            self._data_key = None

        """
        Code for import (deserialization)
        """

        if import_enabled == _NA:
            self.import_enabled = True
        else:
            self.import_enabled = import_enabled

        if self.import_enabled:
            # Default value when key is not found during import
            if import_default == _NA:
                self.import_default = fields.allow_missing
            else:
                self.import_default = import_default
            field_kwargs["missing"] = self.import_default

            # whether to raise error when key is not found during import
            if import_required == _NA:
                self.import_required = False
            else:
                self.import_required = import_required
            field_kwargs["required"] = self.import_required
        else:
            field_kwargs["dump_only"] = True

        """
        Code for export (Serialization)
        """
        if export_enabled == _NA:
            self.export_enabled = True
        else:
            self.export_enabled = export_enabled

        if self.export_enabled:
            # Default value when attribute is not found during export
            if export_default == _NA:
                self.export_default = fields.allow_missing
            else:
                self.export_default = export_default
            field_kwargs["missing"] = self.export_default

            # whether to raise error when key is not found during export
            if export_required == _NA:
                self.export_required = False
            else:
                self.export_required = export_required

        else:
            # TODO: test code behaviors when both load_only and
            #   dump_only are true
            field_kwargs["load_only"] = True

        if doc == _NA:
            doc = None
        self.doc = doc

        # To be used by self._make_field
        self._field_kwargs = field_kwargs

        """
        To be set by __set_name__ when attribute instance is binded
            to a class
        """
        self.parent = _NA
        self.name = _NA

        if requires == _NA:
            requires = list()
        else:
            raise NotImplementedError
        self.requires = requires
        self.type_cls = type_cls

        if is_concrete is _NA:
            is_concrete = False
        self.is_concrete = is_concrete

        if validate is not _NA:
            self._field_kwargs['validate'] = validate


class Boolean(AttributeBase, MakePonyAttribute):

    # def _make_graphql_type(self):
    #     import graphql
    #     return graphql.GraphQLBoolean


    """Field that serializes to a boolean and deserializes
        to a boolean.
    """
    pass

    # def __init__(
    #         self,
    #         *,
    #         initialize,
    #         initialize_value,
    #         data_key,
    #
    #         import_enabled,
    #         import_default,
    #         import_required,
    #
    #         export_enabled,
    #         export_default,
    #         export_required,
    #
    #         type_cls: Optional[Type[T]],):
    #
    #     res = dict()
    #
    #     res["import_only"], res["export_only"] = \
    #         import_enabled and not export_enabled, \
    #         export_enabled and not import_enabled
    #
    #     if import_enabled:
    #         res["required"] = import_required
    #
    #     super().__init__(
    #
    #         value_if_not_loaded=value_if_not_loaded,
    #         nullable=True,
    #         *args,
    #         **kwargs
    #     )
    #
    # def __get__(self, instance, owner) -> bool:
    #     return super().__get__(instance, owner)


class PropertyAttributeBase(AttributeBase):
    """
    Ref: https://blog.csdn.net/weixin_43265804/article/details/82863984
        content under CC 4.0

    TODO: note that later definition of a getter/setter may override earlier
    TODO:   definitions. (this may happen when you subclass a class)

    TODO: check for memory leak
    """

    def  __init__(self,
                  *, fget=None, fset=None, fdel=None,
                  **kwargs):

        """
        Getter
        """

        self.fget = fget

        """
        Setter
        """

        self.fset = fset

        """
        Deleter 
        """

        self.fdel = fdel

        """
        TODO: check that __doc__ is forwarded 
        """

        super().__init__(**kwargs)

    # @typing.overload
    # def __get__(self, instance: typing.Any, owner: typing.Any):
    #     ...

    def _get_select(self, instance):
        return self.fget()

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            fget = self.fget
            if fget is None:
                def fget(_self):
                    inner = getattr(_self, _ATTRIBUTE_STORE_NAME)
                    return getattr(inner, self.name)
            return fget(instance)

    def __set__(self, instance, value):
        if instance is None:
            raise ValueError
        else:
            fset = self.fset
            if fset is None:
                def fset(_self_obj, value):
                    inner = getattr(_self_obj, _ATTRIBUTE_STORE_NAME)
                    return setattr(inner, self.name, value)
            fset(instance, value)

    def __delete__(self, instance):
        if instance is None:
            raise ValueError
        else:
            fdel = self.fdel
            if fdel is None:
                def fdel(_self):
                    inner = getattr(_self, _ATTRIBUTE_STORE_NAME)
                    return delattr(inner, self.name)
            fdel(instance)

    def getter(self, fget):
        _self = self.copy()
        # fget = _self._register_pony_function(fget)
        _self.fget = fget
        return _self

    def setter(self, fset):
        _self = self.copy()
        _self.fset = fset
        return _self

    def deleter(self, fdel):
        _self = self.copy()
        _self.fdel = fdel
        return _self

    def init(self, initializer):
        _self = self.copy()
        _self.initializer = initializer
        return _self


class PropertyAttribute(PropertyAttributeBase, MakePonyAttribute):
    pass


class DictAttribute(PropertyAttribute):

    # def _make_field(self) -> fields.Field:
    #     """
    #     TODO: implement
    #     :return:
    #     """
    #     field_cls: Type[fields.Field] = fields.Dict
    #     return field_cls(**self._field_kwargs, attribute=self.name)

    def __init__(self, **kwargs):

        def _dict_initializer(_self):
            attr_name = self.name
            setattr(_self, attr_name, dict())

        super().__init__(
            initialize=True,
            initializer=_dict_initializer,
            **kwargs)


class RelationshipAttribute(PropertyAttributeBase, MakePonyAttribute):

    def _make_pony_attribute_cls(self):
        from pony.orm.core import Set
        collection_type = self.collection
        if collection_type is not None:
            return Set
        else:
            return super()._make_pony_attribute_cls()

    # @lru_cache(maxsize=None)
    def _make_pony_attribute(self):
        # column_name = self.data_key
        # TODO: Column name has no actual use; so it is not allowed in pony
        py_type = self.dm_cls
        is_required = self.import_required
        reverse = self.reverse if isinstance(self.reverse, str) or self.reverse is None else self.reverse.data_key
        _pony_attribute_cls = self._make_pony_attribute_cls()
        d = dict(is_required=is_required, reverse=reverse)
        if self.data_key != self.name:  # TODO: make better
            d['column'] = self.data_key
        _pony_attribute = _pony_attribute_cls(py_type, **d)
        return _pony_attribute

    def _make_field_cls(self):
        # Makes marshmallow field cls
        return fields.Relationship

    def _make_field(self) -> fields.Field:
        """
        TODO: implement
        :return:
        """
        field_cls = self._make_field_cls()
        one = field_cls(**self._field_kwargs, attribute=self.name)

        if self.collection is None:
            return one
        elif self.collection is dict:
            # TODO: test rigorously; see if the name _Temporary__make_field may
            #     cause it to be affected by another run of _make_field
            dict_field_cls = type(
                '_Temporary__make_field',
                (fields.Mapping,),
                {
                    'mapping_type': self.collection
                })
            return dict_field_cls(values=one)
        elif self.collection is list:
            # TODO: confirm that set and tuple and etc. are compatible
            list_field_cls = fields.List
            return list_field_cls(one)
        else:
            raise NotImplementedError

    def __init__(self, *, nested=_NA, collection=_NA, dm_cls=_NA, reverse=_NA, **kwargs):
        # TODO: compare _NA everywhere with "is" rather than "=="
        super().__init__(
            **kwargs,
        )
        if nested == _NA:
            raise ValueError
        else:
            self.nested = nested
        if self.nested:
            self._field_kwargs["nested"] = self.nested

        if collection == _NA:
            collection = None
        self.collection = collection

        if dm_cls == _NA:
            dm_cls = None
        self.dm_cls = dm_cls
        self._field_kwargs["obj_type"] = self.dm_cls

        if reverse is _NA:
            reverse = None
        self.reverse = reverse

    @property
    def pytype(self):
        if self.collection is None:
            return self.dm_cls
        else:
            raise TypeError  # TODO: make better


class LocalTimeAttribute(PropertyAttribute):

    def _make_field_cls(self):
        # Makes marshmallow field cls
        return fields.Localtime

    def _make_field(self) -> fields.Field:
        field_cls = self._make_field_cls()
        return field_cls(**self._field_kwargs, attribute=self.name)


class StringAttribute(PropertyAttribute):
    def _make_field_cls(self):
        # Makes marshmallow field cls
        return fields.String


class DocRefAttribute(PropertyAttributeBase, MakePonyAttribute):

    def _make_pony_attribute_cls(self):
        return PrimaryKey

    # @lru_cache(maxsize=1)
    def _make_pony_attribute(self):
        column_name = self.data_key
        py_type = self.type_cls
        _pony_attribute_cls = self._make_pony_attribute_cls()
        _pony_attribute = _pony_attribute_cls(
            py_type, column=column_name)
        return _pony_attribute

    def _make_field_cls(self):
        # Makes marshmallow field cls
        return fields.DocRefField

    def _make_field(self) -> fields.Field:
        field_cls = self._make_field_cls()
        return field_cls(**self._field_kwargs, attribute=self.name)


class MultiDocRefAttribute(AttributeBase, MakePonyAttribute):

    def _make_pony_attribute_cls(self):
        return PrimaryKey

    def __init__(self, *attrs):
        self.attributes = list(attrs)
        self.is_concrete = True

    # @lru_cache(maxsize=1)
    def _make_pony_attribute(self):
        _pony_attribute_cls = self._make_pony_attribute_cls()
        _pony_attribute = _pony_attribute_cls(*(attr._pony_attribute for attr in self.attributes))
        return _pony_attribute


    def _make_field_cls(self):
        # Makes marshmallow field cls
        return fields.DocRefField

    def _make_field(self) -> fields.Field:
        field_cls = self._make_field_cls()
        return field_cls(attribute=self.name)


class ReferenceAttribute(PropertyAttribute):

    def _make_field_cls(self):
        # Makes marshmallow field cls
        return fields.StructuralRef

    def _make_field(self) -> fields.Field:
        field_cls = self._make_field_cls()
        return field_cls(**self._field_kwargs, attribute=self.name)

    def __init__(self, many=_NA, dm_cls=_NA, missing=_NA, **kwargs):
        super().__init__(
            **kwargs
        )

        if many == _NA:
            many = False
        self.many = many
        self._field_kwargs["many"] = self.many

        if missing == _NA:
            missing = None
        self.missing = missing
        self._field_kwargs['missing'] = missing

        if dm_cls == _NA:
            dm_cls = None
        self.dm_cls = dm_cls
        self._field_kwargs["obj_type"] = self.dm_cls


class EmbeddedAttribute(PropertyAttribute):

    def _make_field_cls(self):
        # Makes marshmallow field cls
        return fields.Embedded

    def _make_field(self) -> fields.Field:
        field_cls = self._make_field_cls()
        return field_cls(**self._field_kwargs, attribute=self.name)

    def __init__(self, type_cls=_NA, collection=_NA, **kwargs):

        super().__init__(
            type_cls = type_cls,
            **kwargs
        )

        # if many == _NA:
        #     many = False
        # self.many = many
        # self._field_kwargs["many"] = self.many

        if type_cls == _NA:
            type_cls = None

        # self.type_cls = obj_cls
        # self._field_kwargs["obj_type"] = self.dm_cls

        if collection == _NA:
            collection = None
        else:
            self._field_kwargs["many"] = True
        self.collection = collection




class ObjectTypeAttribute(PropertyAttributeBase, TypeClsAsArgMixin):

    def _make_field_cls(self):
        # Makes marshmallow field cls
        return fields.ObjectTypeField

    def _make_field(self) -> fields.Field:
        field_cls = self._make_field_cls()
        return field_cls(**self._field_kwargs, attribute=self.name)

    def __init__(self, f_serialize=_NA, f_deserialize=_NA, **kwargs):
        super().__init__(
            **kwargs,
            # TODO: recover
            # column=self.data_key
        )
        self._field_kwargs["serialize"] = f_serialize
        self._field_kwargs["deserialize"] = f_deserialize

#
# class Attribute(AttributeBase):  # TODO: change
#
#     def __init__(
#             self,
#             *,
#             type_cls=None,
#             **kwargs,
#     ):
#         super().__init__(type_cls=type_cls, **kwargs)
#
#     @typing.overload
#     def __get__(self, instance, owner) -> int:
#         pass
#
#     def __get__(self, instance, owner) -> typing.Any:
#         if instance is None:
#             return self
#         else:
#             return getattr(instance._attribute_store, self.name)
#
#     def __set__(self, instance, value):
#         setattr(instance._attribute_store, self.name, value)
#
#     def __delete__(self, instance):
#         delattr(instance._attribute_store, self.name)


class ForwardInnerAttribute(PropertyAttribute):

    def __init__(self, *, inner_name, **kwargs):

        def fget(_self):
            inner = getattr(_self, inner_name)
            return getattr(inner, self.name)

        def fset(_self, value):
            inner = getattr(_self, inner_name)
            return setattr(inner, self.name, value)

        super().__init__(fget=fget, fset=fset, **kwargs)



import typing
from collections import namedtuple

import iso8601
import pytz
from marshmallow import fields

from onto.common import _NA
from onto.mapper.helpers import RelationshipReference, EmbeddedElement

from datetime import datetime

from marshmallow import utils as mutils

from onto.registry import ModelRegistry

# Firestore Integer supports up to 64-bit signed
# Note that this may result in defect where business data reaches
#   this maximum.
# from onto.models.meta import ModelRegistry

_POS_INF_APPROX = 2 ** 63 - 1
_NEGATIVE_INF_APPROX = -2 ** 63

allow_missing = fields.missing_

OBJ_TYPE_ATTR_NAME = "obj_type"


class Field(fields.Field):
    """
    Custom class of field for supporting onto-related
        features such as auto-initialization.
    """

    def __init__(self,
                 *args, default=fields.missing_,
                 **kwargs):
        super().__init__(*args, default=default, **kwargs)

    @property
    def default_value(self):
        return self.deserialize(fields.missing_)
        # return self.initialize_value() \
        #     if callable(self.initialize_value) \
        #     else self.initialize_value


class Boolean(fields.Bool, Field):
    """Field that serializes to a boolean and deserializes
        to a boolean.
    """

    def __init__(self, missing=bool, *args, **kwargs):
        super().__init__(missing=missing, *args, **kwargs)

    def __get__(self, instance, owner) -> bool:
        return super().__get__(instance, owner)


class NumberTimestamp(fields.Raw, Field):
    pass


class Integer(fields.Integer, Field):
    """Field that serializes to an integer and deserializes
            to an integer.
    """

    def __init__(self, missing=int, *args, **kwargs):
        super().__init__(missing=missing, *args, **kwargs)

    @typing.overload
    def __get__(self, instance, owner) -> typing.Union[Field, int]:
        """
        Type hinting
        """
        pass


class Float(fields.Float, Field):
    """Field that serializes to an float and deserializes
            to an float.
    """

    def __init__(self, missing=float, *args, **kwargs):
        super().__init__(missing=missing, *args, **kwargs)

    @typing.overload
    def __get__(self, instance, owner) -> typing.Union[Field, float]:
        """
        Type hinting
        """
        pass


class Raw(fields.Raw, Field):
    pass


class List(fields.List, Field):
    # TODO: change

    @typing.overload
    def __get__(self, instance, owner) -> typing.Union[Field, typing.List]:
        """
        Type hinting
        # TODO: prevent this section of code from being accidentally called
            even if typing.overload decorator got accidentally erased
        """
        pass

    def __init__(self, *args, missing=list, **kwargs):
        super().__init__(*args, missing=missing, **kwargs)


class Mapping(fields.Mapping, Field):
    def __init__(self, *args, missing=dict, **kwargs):
        super().__init__(*args, missing=missing, **kwargs)


class Dict(fields.Dict, Field):

    @typing.overload
    def __get__(self, instance, owner) -> typing.Union[Field, typing.Dict]:
        """
        Type hinting
        """
        pass

    def __init__(self, *args, missing=dict, **kwargs):
        super().__init__(*args, missing=missing, **kwargs)


class Function(fields.Function, Field):
    """
    For use with property.
    """
    def __init__(self, missing=fields.missing_, *args, **kwargs):
        super().__init__(missing=missing, *args, **kwargs)


class DocIdField(fields.Function, Field):

    def __init__(self, *args, **kwargs):
        def serialize(obj):
            """ serialize: A callable from which to retrieve the value.
            The function must take a single argument ``obj`` which is the object
            to be serialized. It can also optionally take a ``context`` argument,
            which is a dictionary of context variables passed to the serializer.
            If no callable is provided then the ```load_only``` flag will be set

            to True.
            :param obj:
            :return:
            """
            # TODO: delete
            try:
                res = getattr(obj, self.attribute)
            except AttributeError as e:
                return fields.missing_
            else:
                return res

        def deserialize(value):
            """ deserialize: A callable from which to retrieve the value.
            The function must take a single argument ``value`` which is the value
            to be deserialized. It can also optionally take a ``context`` argument,
            which is a dictionary of context variables passed to the deserializer.
            If no callable is provided then ```value``` will be passed through
            unchanged.

            :param value:
            :return:
            """
            return value
            # data_key = self.data_key
            # if data_key not in value:
            #     return fields.missing_
            # else:
            #     return value
            #     # return value[data_key]

        super().__init__(
            *args,
            serialize=serialize,
            deserialize=deserialize,
            **kwargs
        )


from onto.database import Reference


class DocRefField(fields.String, Field):

    def _deserialize(self, value, *args, **kwargs) -> Reference:
        return Reference.from_str(s=value)

    def _serialize(self, value, *args, **kwargs) -> str:
        return str(value)


argument = namedtuple('argument', ['key', 'comparator', 'val'])


class ObjClsMixin:

    def __init__(self, *args, obj_type=_NA, **kwargs):
        if obj_type is _NA:
            from onto.firestore_object import FirestoreObject
            obj_type = FirestoreObject
        self._obj_cls = obj_type
        super().__init__(*args, **kwargs)

    @property
    def obj_cls(self):
        _obj_cls = self._obj_cls
        if isinstance(self._obj_cls, str):
            _obj_cls = ModelRegistry.get_cls_from_name(obj_type_str=_obj_cls)
        return _obj_cls


class ObjectTypeField(fields.Function, Field):

    @staticmethod
    def f_serialize(obj, context):
        """ serialize: A callable from which to retrieve the value.
        The function must take a single argument ``obj`` which is the object
        to be serialized. It can also optionally take a ``context`` argument,
        which is a dictionary of context variables passed to the serializer.
        If no callable is provided then the ```load_only``` flag will be set
        to True.

        :param obj:
        :return:
        """
        # TODO: delete

        try:
            from onto.utils import obj_type_serialize
            res = obj_type_serialize(obj)
        except AttributeError:
            return fields.missing_
        else:
            return res

    def read_obj_type_str(self, raw_dict) -> typing.Optional[str]:
        """ Read obj_type string from inputs that are to be
        deserialized/imported/loaded soon

        :return:
        """
        if self.data_key in raw_dict:
            return raw_dict[self.data_key]
        else:
            return None

    def get_obj_type_data_key(self):
        # NOTE: this is different from not self.dump_only
        # NOTE: load_only and dump_only can be true at the same time
        if self.load_only:
            return None
        else:
            return self.data_key

    def get_obj_type_condition(self, obj_cls) -> typing.Optional[typing.Tuple]:
        data_key = self.get_obj_type_data_key()
        if data_key is None:
            return None
        else:
            return argument(
                key=data_key,
                comparator=obj_cls._datastore().Comparators._in,
                val=obj_cls._get_subclasses_str())

    def __init__(self, *args, serialize=_NA, data_key="obj_type",
                 deserialize=_NA,
                 **kwargs):
        if serialize is _NA:
            serialize = self.f_serialize
        if deserialize is _NA:
            deserialize = None

        super().__init__(
            *args,
            serialize=serialize,
            data_key=data_key,
            deserialize=deserialize,
            **kwargs
        )


class String(fields.String, Field):

    @typing.overload
    def __get__(self, instance, owner) -> typing.Union[Field, str]:
        """
        Type hinting
        """
        pass

    def __init__(self, missing=str, *args, **kwargs):
        super().__init__(missing=missing, *args, **kwargs)


class Nested(fields.Nested, Field):
    """
    Field that describes a dictionary that conforms to a marshmallow schema.
    """
    pass


class _MissingNotSpecified:
    pass


class Relationship(ObjClsMixin, fields.Str, Field):
    """
    Field that describes a relationship in reference to another document
        in the Firestore.
    """

    def __init__(self, *args, missing=_MissingNotSpecified, nested=False,
                 many=False, **kwargs):
        """ Initializes a relationship. A field of the master object
                to describe relationship to another object or document
                being referenced. Set missing=dict if many=True and
                data is stored in a dict.

        :param args: Positional arguments to pass to marshmallow.fields.Str
        :param nested: If set to True, the document being referenced
                    will be retrieved and saved as the master document
                    or object. If set to False, only the reference
                    (DocumentReference) will be stored in the master
                    document and retrieved into the master object.
        :param many: If set to True, will deserialize and serialize the field
                    as a list. (TODO: add support for more iterables)
        :param kwargs: Keyword arguments to pass to marshmallow.fields.Str
        """
        if missing == _MissingNotSpecified:
            missing = list if many else None
        super().__init__(*args, missing=missing, **kwargs)
        self.nested = nested
        self.many = many

    def _serialize(self, value, *args, **kwargs):
        if value is None:
            return None
            # raise ValueError

        if isinstance(value, list) and self.many:
            return [self._serialize(val, *args, **kwargs) for val in value]
        elif isinstance(value, dict) and self.many:
            val_d = dict()
            for k, v in value.items():
                val_d[k] = self._serialize(v, *args, **kwargs)
            return val_d

        if not self.nested:
            # Note that AssertionError is not always thrown
            return RelationshipReference(doc_ref=str(value), nested=self.nested)
        else:
            return RelationshipReference(obj=value, nested=self.nested)

    def _deserialize(self, value, *args, **kwargs):
        if value is None:
            return None
            # raise ValueError

        if isinstance(value, list) and self.many:
            return [self._deserialize(val, *args, *kwargs) for val in value]
        elif isinstance(value, dict) and self.many:
            val_d = dict()
            for k, v in value.items():
                val_d[k] = self._deserialize(v, *args, **kwargs)
            return val_d

        # assert isinstance(value, DocumentReference)
        from onto.database.firestore import FirestoreReference
        if isinstance(value, str):
            doc_ref = FirestoreReference.from_str(value)
        else:
            doc_ref = FirestoreReference.from_document_reference(value)
        # TODO: change
        return RelationshipReference(
            doc_ref=doc_ref,
            nested=self.nested,
            obj_type=self.obj_cls
        )


class StructuralRef(ObjClsMixin, fields.Str, Field):

    def __init__(self, *args, missing=_MissingNotSpecified,
                 many=False, **kwargs):
        """ Initializes a relationship. A field of the master object
                to describe relationship to another object or document
                being referenced.

        :param args: Positional arguments to pass to marshmallow.fields.Str
        :param many: If set to True, will deserialize and serialize the field
                    as a dict. (TODO: add support for more iterables)
        :param kwargs: Keyword arguments to pass to marshmallow.fields.Str
        """
        if missing == _MissingNotSpecified:
            missing = dict if many else None
        super().__init__(*args, missing=missing, **kwargs)
        self.many = many

    def _serialize(self, value, *args, **kwargs):
        if value is None:
            return None
            # raise ValueError

        if isinstance(value, list) and self.many:
            return [self._serialize(val, *args, **kwargs) for val in value]
        elif isinstance(value, dict) and self.many:
            val_d = dict()
            for k, v in value.items():
                val_d[k] = self._serialize(v, *args, **kwargs)
            return val_d

        # if isinstance(value, DocumentReference):
        #     # Note that AssertionError is not always thrown
        #     return RelationshipReference(doc_ref=value, nested=True)
        # else:
        return RelationshipReference(obj=value, nested=True)

    def _deserialize(self, value, *args, **kwargs):
        if value is None:
            return None
            # raise ValueError
        from onto.store.struct import struct_ref

        if isinstance(value, list) and self.many:

            return [self._deserialize(val, *args, *kwargs) for val in value]

        elif isinstance(value, dict) and self.many:

            val_d = dict()
            for k, v in value.items():
                val_d[k] = self._deserialize(v, *args, **kwargs)
            return val_d

        elif isinstance(value, struct_ref):

            if value.snapshot is not None and value.ref is not None:
                dm_cls, snapshot, ref = value.dm_cls, value.snapshot, value.ref
                obj = dm_cls.from_snapshot(ref=ref, snapshot=snapshot)
                return RelationshipReference(
                    obj=obj,
                    nested=True,
                    obj_type=dm_cls
                )
            elif value.obj is not None:
                obj = value.obj
                return RelationshipReference(
                    obj=obj,
                    nested=True,
                    obj_type=obj.__class__
                )
            elif value.doc_ref is not None:
                dm_cls, doc_ref = value.dm_cls, value.ref
                return RelationshipReference(
                    doc_ref=doc_ref,
                    nested=True,
                    obj_type=dm_cls
                )
            elif value.doc_id is not None:
                dm_cls, doc_id = value.dm_cls, value.id
                doc_ref = self.obj_cls.ref_from_id(doc_id=doc_id)
                return RelationshipReference(
                    doc_ref=doc_ref,
                    nested=True,
                    obj_type=dm_cls
                )
            else:
                raise ValueError

        else:

            dm_cls, doc_id = value
            doc_ref = self.obj_cls.ref_from_id(doc_id=doc_id)
            return RelationshipReference(
                doc_ref=doc_ref,
                nested=True,
                obj_type=dm_cls
            )


class Embedded(ObjClsMixin, fields.Raw, Field):
    """
    Note that when many is set to True, default value of this field
        is an empty list (even if a dict is expected).
    """

    def __init__(self, *args, missing=_MissingNotSpecified, many=False,
                 **kwargs):
        """

        :param args: Positional arguments to pass to marshmallow.fields.Str
        :param many: If set to True, will deserialize and serialize the field
                    as a list. (TODO: add support for more iterables)
        :param obj_cls: cls of the element
        :param kwargs: Keyword arguments to pass to marshmallow.fields.Str
        """

        if missing == _MissingNotSpecified:
            missing = list if many else None
        self.many = many

        super().__init__(*args, missing=missing, **kwargs)

    def _serialize(self, value, *args, embed_many=None, **kwargs):
        if embed_many is None:
            embed_many = self.many

        if embed_many:
            if isinstance(value, list):
                return [self._serialize(val, embed_many=False)
                        for val in value]
            elif isinstance(value, dict):
                return {
                    key: self._serialize(val, embed_many=False)
                    for key, val in value.items()
                }
            else:
                raise NotImplementedError
        else:
            return EmbeddedElement(
                obj=value
            )

    def _deserialize(self, value, *args, embed_many=None, **kwargs):
        if embed_many is None:
            embed_many = self.many

        if embed_many:
            if isinstance(value, list):
                return [self._deserialize(
                    val, *args, **kwargs, embed_many=False)
                    for val in value]
            elif isinstance(value, dict):
                return {
                    key: self._deserialize(
                        val, *args, **kwargs, embed_many=False)
                    for key, val in value.items()
                }
            else:
                raise NotImplementedError
        else:
            return EmbeddedElement(
                d=value,
                obj_cls=self.obj_cls
            )


def local_time_from_timestamp(timestamp) -> datetime:
    """
    Accurate to seconds
    :param timestamp: for example: 1545062400
    :return: for example: "2018-12-17T08:00:00"
    """
    tz = pytz.timezone('US/Pacific')  # ('America/Los_Angeles')   // TODO: make consistent with str_to_local_time

    d: datetime = datetime.fromtimestamp(timestamp, tz=tz)
    d = d.replace(tzinfo=None)  # Convert to local time
    return d.isoformat()


def str_to_local_time(s) -> datetime:
    tz = pytz.timezone('America/Los_Angeles')
    return tz.localize(iso8601.parse_date(s, default_timezone=None))


def timestamp_from_local_time(s) -> int:
    return int(str_to_local_time(s).timestamp())


class Localtime(fields.NaiveDateTime, Field):

    @typing.overload
    def __get__(self, instance, owner) -> typing.Union[Field, int]:
        """
        Type hinting
        """
        pass

    def _serialize(
            self, value, *args, **kwargs
    ):
        if value is None:
            return None
        else:
            return local_time_from_timestamp(value)

    def _deserialize(self, value, *args, **kwargs):
        if value is None:
            return mutils.missing
        else:
            return timestamp_from_local_time(value)


class Remainder(fields.Dict, Field):
    """
    To match fields that are not declared.
    """

    def __init__(self, missing=fields.missing_, *args, **kwargs):
        super().__init__(missing=missing, *args, **kwargs)


# class BpStoreField(fields.Raw, Field):
#
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, data_key="_structure", **kwargs)
#         self.many = many


Str = String



import pytest
from flask import Flask

from onto import testing_utils
from .fixtures import CTX
from .color_fixtures import color_refs, ColorSchema, ColorDomainModelBase, \
    Color, rainbow_vm
from tests.fixtures import setup_app
from unittest import mock


def test_rainbow_stuffs(CTX, setup_app, color_refs, rainbow_vm):

    RainbowViewModelDAV = rainbow_vm

    def notify(obj):
        obj.save()

    obj = RainbowViewModelDAV.get("yellow+magenta+cian",
                                  f_notify=notify)  # Non-standard usage

    vm_id = obj.doc_ref.id

    assert CTX.db.collection("RainbowDAV").document(vm_id).get().to_dict() == {
        'colors': ['cian', 'magenta', 'yellow'],
        'rainbowName': 'cian-magenta-yellow'
    }



from onto.common import _NA
from onto.database import Database, Reference, Snapshot
from onto.context import Context as CTX

class CouchDatabase(Database):
    import couchdb

    @classmethod
    def server(cls) -> couchdb.Server:
        return CTX.dbs.couch

    @classmethod
    def set(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):

        if transaction is not _NA:
            raise ValueError

        k = ref.first
        _id = ref.last

        db = cls.server()[k]
        db.save({
            '_id': _id,
            **snapshot
        })



from google.cloud.firestore_v1 import Query

from examples.fluttergram.domain_models import Post, User
from onto.store.business_property_store import BPSchema
from onto.store.struct import Struct

from onto.view import ViewMediatorDeltaDAV
from onto import utils, schema, fields, testing_utils
from onto import view_model
# Import the fixtures used by fixtures


class PostDAVSchema(schema.Schema):
    post = fields.Embedded(dump_only=True)


class PostStoreBpss(BPSchema):
    consumer = fields.StructuralRef(dm_cls=User)
    post = fields.StructuralRef(dm_cls=Post)


class PostDAV(view_model.ViewModel):
    class Meta:
        schema_cls = PostDAVSchema

    @property
    def post(self):
        return self.store.post

    @classmethod
    def new(cls, *args, consumer_id, post_id, **kwargs):
        ref = User._get_collection().document(consumer_id).collection("feed") \
            .document(post_id)

        struct = Struct(schema_obj=PostStoreBpss())
        struct["consumer"] = (User, consumer_id)
        struct["post"] = (Post, post_id)

        obj = cls.get(*args, struct_d=struct, doc_ref=ref, **kwargs)

        return obj


class FeedEntryViewMediatorDeltaDAV(ViewMediatorDeltaDAV):

    def __init__(self, *args, post_cls=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.post_cls = post_cls

    def _get_query_and_on_snapshot(self):
        query = Query(parent=self.post_cls._get_collection())

        def on_snapshot(snapshots, changes, timestamp):
            for change, snapshot in zip(changes, snapshots):
                if change.type.name == 'ADDED':
                    post = utils.snapshot_to_obj(
                        snapshot,
                        super_cls=self.post_cls
                    )

                    producer_id = post.owner_id
                    producer = User.get(doc_id=producer_id)
                    followers = list()
                    for key, val in producer.followers.items():
                        if val is True:
                            followers.append(key)

                    for consumer_id in followers:
                        obj = self.view_model_cls.new(
                            consumer_id=consumer_id,
                            post_id=post.doc_id,
                            once=False,
                            f_notify=self.notify
                        )

        return query, on_snapshot


def test_start(users, posts):
    mediator = FeedEntryViewMediatorDeltaDAV(
        view_model_cls=PostDAV,
        post_cls=Post,
    )

    mediator.start()

    testing_utils._wait(factor=.2)

    ref = User._get_collection().document("thomasina").collection("feed") \
        .document(posts[0].doc_id)

    assert ref.get().to_dict() == {'obj_type': 'PostDAV',
                                   'post': {'ownerId': 'tijuana',
                                            'doc_ref': 'insta_posts/tj_t',
                                            'obj_type': 'Post',
                                            'hello': 'world',
                                            'doc_id': 'tj_t'},
                                   'doc_ref': 'users/thomasina/feed/tj_t'}



import functools

from onto.common import _NA
from onto.database import Database, Reference, Snapshot, Listener
from onto.context import Context as CTX

import kafka

# TODO: NOTE maximum of 1 firestore client allowed since we used a global var.
from typing import List
from onto.query.query import Query
from onto.store.snapshot_container import SnapshotContainer
from math import inf

FirestoreListener = None


class KafkaReference(Reference):

    def is_collection(self):
        return len(self.params) % 2 == 1

    @property
    def collection(self):
        return self.first

    def is_document(self):
        return len(self.params) % 2 == 0

    @classmethod
    def from__document_name(cls, document_name: str):
        return cls.from_str(document_name)

    @property
    def _document_path(self):
        return str(self)


class KafkaReadDatabase(Database):

    @classmethod
    def listener(cls):
        from onto.database.utils import GenericListener
        return GenericListener

    bootstrap_servers = None

    d = dict()

    @classmethod
    def _onto_set(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
        cls.d[str(ref)] = snapshot.to_dict()
        cls.listener()._pub(reference=ref, snapshot=snapshot)

    @classmethod
    def get(cls, ref: Reference, transaction=_NA):
        return Snapshot(cls.d[str(ref)])

    update = set
    create = set

    @classmethod
    def delete(cls, ref: Reference, transaction=_NA):
        """ Note: this only deletes one instance that has _doc_id == ref.last

        :param ref:
        :param transaction:
        :return:
        """
        del cls.d[str(ref)]


class KafkaSnapshot(Snapshot):

    @classmethod
    def from_data_and_meta(
            cls, **kwargs):
        """

        :param kwargs:
        :return:
        """
        DATA_KEYWORD = 'data'
        if DATA_KEYWORD not in kwargs:
            raise ValueError
        else:
            data = kwargs[DATA_KEYWORD]
            __onto_meta__ = {
                key: val
                for key, val in kwargs.items()
                if key != DATA_KEYWORD
            }
            return cls(**data, __onto_meta__=__onto_meta__)

    @classmethod
    def empty(cls, **kwargs):
        return cls.from_data_and_meta(
            data=dict(),
            exists=False,
            **kwargs
        )




from flask import Flask, Blueprint

from .base import Source

from flask_jsonrpc import JSONRPC


class JsonRpcSource(Source):

    def __init__(self, *args, url_prefix, **kwargs):
        super().__init__(*args, **kwargs)
        self.url_prefix = url_prefix

    def start(self, app: Flask):
        jsonrpc = JSONRPC(app, self.url_prefix, enable_web_browsable_api=True)
        for rule, f_name in self.protocol.mapping.items():
            wrapper = jsonrpc.method(name=rule, validate=False)
            @wrapper
            def routable(*args, **kwargs):
                return self._invoke_mediator(func_name=rule, *args,  **kwargs)



from flask import Response, jsonify

from examples.meeting_room.domain_models import Meeting
from examples.meeting_room.view_models import MeetingSession
from examples.meeting_room.view_models.meeting_session import MeetingSessionC
from onto.source.rest import RestViewModelSource
from onto.view import Mediator

# class RestMediator(Mediator):
#
#     view_model_cls = None
#     source = RestViewModelSource()


class ViewModelResponse(Response):
    pass


class MeetingSessionRest(Mediator):

    # from onto import source, sink

    view_model_cls = MeetingSessionC

    rest = RestViewModelSource()

    @rest.route('/<doc_id>', methods=('GET',))
    def materialize_meeting_session(self, doc_id):

        meeting = Meeting.get(doc_id=doc_id)

        def notify(obj):
            d = obj.to_snapshot().to_dict()
            content = jsonify(d)
            self.rest.emit(content)

        _ = MeetingSessionC.get(
            doc_id=meeting.doc_id,
            once=False,
            f_notify=notify
        )

    # @rest.route('/', methods=('GET',))
    # def list_meeting_ids(self):
    #     return [meeting.to_snapshot().to_dict() for meeting in Meeting.all()]

    @classmethod
    def start(cls, app):
        cls.rest.start(app)



import enum
import typing
from collections import namedtuple
from functools import partial, lru_cache

from onto.attrs.unit import Monad, MonadPrependMixin

try:
    from pony.orm.core import Attribute, PrimaryKey, Discriminator, Optional, Required
except ImportError:
    import warnings
    warnings.warn("pony import skipped")

from onto.mapper import fields
from typing import Type, Callable

from onto.common import _NA
from onto.query.cmp import Condition, RootCondition

from functools import cached_property

_ATTRIBUTE_STORE_NAME = "_attrs"

class ValueNotProvided:
    pass


class TypeClsAsArgMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class MakePonyAttribute:

    def _make_pony_attribute_cls(self):
        if self.__class__.__name__ == "Discriminator":
            # TODO: make better
            return Discriminator
        if self.import_required:
            return Required
        else:
            return Optional

    # @lru_cache(maxsize=None)
    def _make_pony_attribute(self):
        column_name = self.data_key
        py_type = self.type_cls
        is_required = self.import_required
        _pony_attribute_cls = self._make_pony_attribute_cls()
        _pony_attribute = _pony_attribute_cls(
            py_type, is_required=is_required, column=column_name)
        return _pony_attribute

    @cached_property
    def _pony_attribute(self):
        if self.is_concrete:
            return self._make_pony_attribute()
        else:
            return None

    @property
    def is_collection(self):
        return self._pony_attribute.is_collection


udf = namedtuple('udf', ['f', 'fid', 'd', 'result_pytype'])


class MakePonyFunction:

    assigned_ids = set()

    @classmethod
    def _random_func_id(cls, N=5):
        import random, string
        fid = '_onto_udf_' + ''.join(random.choices(string.ascii_lowercase, k=N))
        if fid in cls.assigned_ids:
            raise ValueError("Collision")
        cls.assigned_ids.add(fid)
        return fid

    @classmethod
    def register_function(cls, f, type_cls):
        class Some:
            pass

        # TODO: add notes about `typing.*` not supported

        # import ast
        # code = f.__code__
        import inspect
        d = list(inspect.getclosurevars(f).unbound)

        def k(*args):
            _self = Some()
            for idx in range(len(d)):
                k = d[idx]
                v = args[idx]
                setattr(_self, k, v)
            _f = f
            from types import MethodType
            _f = MethodType(f, _self)
            return _f()

        fid = cls._random_func_id()
        udf_obj = udf(f=k, fid=fid, d=d, result_pytype=type_cls)
        # f_reg.append(udf_obj)
        return udf_obj

    @staticmethod
    def _flink_type_mapping(type_cls):
        from pyflink.table.types import _type_mappings
        return _type_mappings[type_cls]

    @classmethod
    def _register_flink_function(cls, udf_obj: udf):
        from pyflink.table import ScalarFunction, DataTypes
        from pyflink.table import udf as flink_udf
        F = type(udf_obj.fid, bases=(ScalarFunction,), dict={
            'eval': udf_obj.f,
        })
        table_env.create_temporary_function(
            udf_obj.fid,
            flink_udf(
                F(),
                input_types=[cls._flink_type_mapping(v) for _, v in udf_obj.d.items()],  # TODO: ensure order of keys
                result_type=cls._flink_type_mapping(udf_obj.result_pytype)
            )
        )
        # flink_udf()

#
# @db.on_connect(provider='flink')
# def _add_udf(db, connection):
#     # cursor = connection.cursor()
#
#     def create_function(name, num_params, func):
#         from pony.orm.dbproviders.sqlite import keep_exception
#         func = keep_exception(func)
#         connection.create_function(name, num_params, func)
#
#     for udf in f_reg:
#         create_function(udf.fid, len(udf.d), udf.f)


class AttributeMixin(RootCondition, Monad, MonadPrependMixin):

    def __hash__(self):
        # TODO: debug
        return id(self)

    def _make_field(self) -> fields.Field:
        """
        TODO: implement
        :return:
        """
        raise NotImplementedError
        # return self.marshmallow_field
        # field_cls = self._make_field_cls()
        # return field_cls(**self._field_kwargs, attribute=self.name)

    def _get_data_key(self):
        """
        TODO: deprecate
        :return:
        """
        return self.data_key

    # def __set_name__(self, owner, name):
    #     self.parent = owner
    #     self.name = name

    @property
    def name(self):
        return self.properties.name

    @property
    def parent(self):
        return self.properties.parent

    def copy(self):
        """
        TODO: debug
        NOTE: contents in self.field_kwargs may or may not be copied
        TODO: fix
        :return:
        """
        from copy import deepcopy
        return deepcopy(self)

    def _get_select(self, instance):
        return getattr(instance, self.properties.name)

    def __get__(self, instance, owner):
        """ Only allow attribute object to be invoked "get" on
                a class, and not an instance.

        :param instance:
        :param owner:
        :return:
        """
        if instance is None:
            return self
        else:
            raise AttributeError()

    # def __init__(
    #         self,
    #         *,
    #         initialize,
    #         initialize_value,
    #         data_key,
    #
    #         import_enabled,
    #         import_default,
    #         import_required,
    #
    #         export_enabled,
    #         export_default,
    #         export_required,
    #
    #         type_cls: Optional[Type[T]],):
    #
    #     res = dict()
    #
    #     res["import_only"], res["export_only"] = \
    #         import_enabled and not export_enabled, \
    #         export_enabled and not import_enabled
    #
    #     if import_enabled:
    #         res["required"] = import_required
    #
    #     super().__init__(
    #
    #         value_if_not_loaded=value_if_not_loaded,
    #         nullable=True,
    #         *args,
    #         **kwargs
    #     )
    #
    # def __get__(self, instance, owner) -> bool:
    #     return super().__get__(instance, owner)


class PropertyMixin:
    """
    Ref: https://blog.csdn.net/weixin_43265804/article/details/82863984
        content under CC 4.0

    """

    # @typing.overload
    # def __get__(self, instance: typing.Any, owner: typing.Any):
    #     ...

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            fget = getattr(self.properties, 'fget')
            return fget(instance)

    def __set__(self, instance, value):
        if instance is None:
            raise ValueError
        else:
            fset = getattr(self.properties, 'fset')
            fset(instance, value)

    def __delete__(self, instance):
        if instance is None:
            raise ValueError
        else:
            fdel = getattr(self.properties, 'fdel')
            fdel(instance)


class AttributeBase(PropertyMixin, AttributeMixin):

    @property
    def marshmallow_field(self):
        try:
            return self.properties._marshmallow_field_constructor(self.properties)
        except Exception as e:
            raise ValueError(f'Failed creating marshmallow_field ') from e

    def _graphql_field_constructor(self):
        arguments = dict(self.properties._graphql_field_kwargs)
        # args = [arguments['type_']]
        # del arguments['type_']
        import graphql
        if not self.properties.is_input:
            from functools import partial
            field_base = graphql.GraphQLField

            from graphql import GraphQLResolveInfo
            def resolve_info(obj, context: GraphQLResolveInfo, **kwargs):
                value = getattr(obj, self.properties.name, None)
                if isinstance(value, enum.Enum):
                    value = value.value
                elif isinstance(value, Callable):
                    f = value
                    value = f(obj, **kwargs)
                return value

            # def resolver(attributed, resolve_info):
            #     return attributed.graphql_field_resolve(resolve_info)
                # raise ValueError
            from onto.helpers.graphql import default_field_resolver

            field_base = partial(
                field_base,
                resolve=resolve_info)
        else:
            field_base = graphql.GraphQLInputField
        field = field_base(
            # *args,
            **arguments
        )
        return field

    @property
    def graphql_field(self):
        try:
            return self._graphql_field_constructor()
        except Exception as e:
            raise ValueError(f'Failed creating graphql for {self.name}') from e

    @property
    def annotation(self):
        if hasattr(self, 'parent'):
            owner = self.parent
            name = self.name
            annotations = typing.get_type_hints(owner)
            # annotations = getattr(owner, '__annotations__', None)
            if annotations is not None:
                if name in annotations:
                    return annotations[name]



from flask import Blueprint, Flask, abort, Response, make_response, jsonify

from onto.source.base import Source
from onto.source.protocol import Protocol


class RestProtocol(Protocol):
    """
    TODO: remove self.mapping initialized in superclass
    """

    def _register(self, rule, methods):
        def decorator(f):
            self.mapping[(rule, methods)] = f.__name__
            return f
        return decorator

    def fname_of(self, func_name):
        return func_name

    def route(self, *args, **kwargs):
        return self._register(*args, **kwargs)

    def __getattr__(self, item):
        raise AttributeError

    def _get_functions(self):
        for (rule, methods), f_name in self.mapping.items():
            yield rule, methods, f_name


class RestSource(Source):

    _protocol_cls = RestProtocol

    def route(self, *args, **kwargs):
        return self.protocol.route(*args, **kwargs)
        # protocol: RestProtocol = self.protocol

    def __init__(self, view_model_cls=None, *args, **kwargs):
        self._mediator_instance = None
        self._view_model_cls = view_model_cls
        super().__init__(*args, **kwargs)

    def emit(self, res):
        abort(res)

    @property
    def mediator_instance(self):
        if self._mediator_instance is None:
            self._mediator_instance = self.parent()()
        return self._mediator_instance

    def start(self, app: Flask, url_prefix=None):
        bp = Blueprint(self.parent().__name__, __name__, url_prefix=url_prefix)
        for rule, methods, f_name in self.protocol._get_functions():
            def routable(*args, **kwargs):
                return self._invoke_mediator(func_name=f_name, *args,  **kwargs)
            bp.add_url_rule(rule=rule, view_func=routable, methods=methods)
        app.register_blueprint(bp)


class RestViewModelSource(RestSource):

    @property
    def view_model_cls(self):
        if self._view_model_cls is not None:
            return self._view_model_cls
        else:
            return self.mediator_instance.view_model_cls

    def start(self, *args, url_prefix=None, **kwargs):
        if url_prefix is None:
            url_prefix = '/' + self.view_model_cls._get_collection_name()
        return super().start(*args, url_prefix, **kwargs)




from onto import domain_model, attrs
# from . import User
# from onto.collection_mixin import with_pony


class TicketBase(domain_model.DomainModel,
                 # metaclass=with_pony
                 ):
    doc_ref = attrs.doc_ref(type_cls=str)

    class Meta:
        collection_name = "tickets"


class Ticket(TicketBase):

    @classmethod
    def _datastore(cls):
        from onto.database.kafka import KafkaDatabase
        return KafkaDatabase

    role = attrs.bproperty(type_cls=str)
    user = attrs.relation(nested=False, dm_cls='User')
    attendance = attrs.bproperty(type_cls=bool)

    meetings = attrs.relation(import_required=False, dm_cls='Meeting', collection=list, nested=False)




import pytest

from onto.attrs import attrs
from .fixtures import CTX


def test_deserialize_doc_id(CTX):

    from onto.domain_model import DomainModel

    class MyDomainModel(DomainModel):

        foo = attrs.string

    obj = MyDomainModel.new(
        doc_id='a',
        foo='bar'
    )

    assert obj.doc_id == 'a'

    d = obj.to_dict()
    assert d['doc_id'] == 'a'

    obj_new = MyDomainModel.from_dict(
        d
    )

    assert obj_new.doc_id == 'a'




from unittest import mock

import pytest
from onto.attrs import attribute
from onto.attrs.attribute import PropertyAttribute, \
    ForwardInnerAttribute, Attribute, AttributeBase

from onto.models.base import Serializable


def test_binding():
    class A(object):
        i = attribute.AttributeBase()

    assert isinstance(A.i, attribute.AttributeBase)

    with pytest.raises(AttributeError):
        _ = A().i


class A(object):

    i = PropertyAttribute()

    @i.getter
    def i(self) -> int:
        return self._i

    @i.setter
    def i(self, value):
        self._i = value


def test_property():
    a = A()
    a.i = 1
    assert a.i == 1
    assert isinstance(a.i, int)
    assert isinstance(A.i, PropertyAttribute)


class B(object):

    i: int = ForwardInnerAttribute(inner_name="aa")


def test_forward_inner():

    b = B()
    a = A()
    a.i = 1
    b.aa = a
    assert b.i == a.i == 1


class C(Serializable):

    class Meta:
        pass

    i = PropertyAttribute()


def test_property_attribute():

    c = C()
    c.i = 2
    assert c.i == 2





def test_initializer():

    f = mock.Mock(return_value=None, __name__='i')

    class D(Serializable):
        class Meta:
            pass

        i = PropertyAttribute(initialize=True)
        # i = PropertyAttribute()

        # i = i.init(f)

        @i.init
        def i(self):
            f()

    d = D()
    d.i = 2
    assert d.i == 2
    f.assert_called_once()


class E(Serializable):

    class Meta:
        pass

    i = PropertyAttribute(initialize=False)

    i = i.init(mock.Mock(return_value=None, __name__='i'))


def test_initializer_not_called():

    e = E()
    e.i = 2
    assert e.i == 2
    E.i.initializer.assert_not_called()


def test_copy():

    a = E.i
    b = E.i.copy()
    assert a is not b


def test_import_only():

    class K(Serializable):
        class Meta:
            import_only = True

        i = PropertyAttribute(initialize=False)

        i = i.init(mock.Mock(return_value=None, __name__='i'))

    k = K.new(i=1)
    assert 'i' not in k.to_dict()



from onto.common import _NA
from onto.database import Database, Reference, Snapshot
from onto.database.utils import GenericListener


class MockReference(Reference):

    def is_collection(self):
        return len(self.params) % 2 == 1

    @property
    def collection(self):
        return self.first

    def is_document(self):
        return len(self.params) % 2 == 0


class MockDatabase(Database):

    class Comparators(Database.Comparators):

        eq = lambda a, b: a == b
        gt = lambda a, b: a > b
        ge = lambda a, b: a >= b
        lt = lambda a, b: a < b
        le = lambda a, b: a <= b
        contains = lambda a, b: a.has(b)  # TODO; check
        _in = lambda a, b: a in b

    @classmethod
    def listener(cls):
        return GenericListener

    d = dict()

    ref = MockReference()

    @classmethod
    def set(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
        cls.d[str(ref)] = snapshot.to_dict()
        cls.listener()._pub(reference=ref, snapshot=snapshot)

    @classmethod
    def get(cls, ref: Reference, transaction=_NA):
        return Snapshot(cls.d[str(ref)])

    update = set
    create = set

    @classmethod
    def delete(cls, ref: Reference, transaction=_NA):
        """ Note: this only deletes one instance that has _doc_id == ref.last

        :param ref:
        :param transaction:
        :return:
        """
        del cls.d[str(ref)]
        cls.listener()._pub(reference=ref, snapshot=None)


    @classmethod
    def query(cls, q):
        qualifier = q._to_qualifier()
        for k, v in cls.d.items():
            if qualifier(v):
                yield MockReference.from_str(k), Snapshot(v)
        yield from ()





import asyncio
import weakref
from collections import defaultdict, OrderedDict
from functools import partial
from inspect import Parameter
from typing import Type

from onto.helpers import make_variable
from onto.source.protocol import Protocol
from onto.view_model import ViewModel

from onto.sink.base import Sink


from collections import namedtuple

graph_schema = namedtuple('graph_schema', ['op_type', 'name', 'graphql_object_type', 'args'])
import graphql

graphql_context = make_variable('graphql_context', default=None)


def get_info() -> 'GraphQLResolveInfo':
    return graphql_context.get()


def get_user(info: 'GraphQLResolveInfo'):
    return info.context['request'].user

class GraphQLSink(Sink):

    _protocol_cls = Protocol

    def __set_name__(self, owner, name):
        """ Keeps mediator as a weakref to protect garbage collection
        Note: mediator may be destructed if not maintained or referenced
            by another variable. Mediator needs to be explicitly kept alive.

        TODO: note that parent and sink_name are not initialized

        :param owner:
        :param name:
        :return:
        """
        self.parent = weakref.ref(owner)
        self._name = name

    @property
    def sink_name(self):
        name = self._name
        if self._camelize:
            from onto.utils import camel
            name = camel(name)
        return name

    @property
    def triggers(self):
        return self.protocol

    @property
    def mediator_instance(self):
        return self.parent()()

    def _maybe_deserialize(self, val, annotated_type):
        from onto.models.base import BaseRegisteredModel
        if issubclass(annotated_type, BaseRegisteredModel):
            if isinstance(val, annotated_type):
                return val  # user: User already deserialized
            return annotated_type.from_dict_special(val)
        else:
            return val

    def _f_of_rule(self, func_name):
        fname = self.protocol.fname_of(func_name)
        if fname is None:
            raise ValueError(
                f"fail to locate {func_name}"
                f" for {self.mediator_instance.__class__.__name__}"
            )
        f = getattr(self.mediator_instance, fname)
        return f

    def _invoke_mediator(self, *args, func_name, **kwargs):

        async def trivial():
            """
            A trivial function to prevent calling await on None
            """
            return None

        try:
            f = self._f_of_rule(func_name=func_name)
        except ValueError as e:
            import logging
            logging.exception('mediator not located')
            return trivial()

        annotation_d = {
            k: v
            for k, v in self._parameters_for(f)
        }

        new_kwargs = {
            k:self._maybe_deserialize(val=v, annotated_type=annotation_d[k]) if k in annotation_d else v
            for k, v in kwargs.items()
        }

        try:
            return f(*args, **new_kwargs)
        except Exception as e:
            import logging
            logging.exception('_invoke mediator failed for graphql subscription')

            return trivial() # TODO: return something else

    def __init__(self, view_model_cls: Type[ViewModel], camelize=True, many=False):
        """

        :param view_model_cls:
        :param camelize:
        """
        self.protocol = self._protocol_cls()
        self.view_model_cls = view_model_cls
        import asyncio
        loop = asyncio.get_event_loop()
        from functools import partial
        cons = partial(asyncio.Queue
                       # , loop=loop
                       )

        from onto.sink.utils import Broker
        self.broker = Broker()
        self.loop = loop
        self._camelize = camelize
        self.many = many
        super().__init__()

    @property
    def qs(self):
        raise TypeError(
            """
            qs is deprecated.
            instead of ``` self.sink.qs[topic_name] = item ```, use:
            ```await self.sink.publish(topic_name, item) ```
            """
        )

    @staticmethod
    def _param_to_graphql_arg(annotated_type):
        from onto.models.utils import _graphql_type_from_py
        from graphql import GraphQLArgument, GraphQLInputObjectType
        # TODO: add notes about `typing.*` not supported
        gql_field_cls = _graphql_type_from_py(annotated_type, input=True)
        arg = GraphQLArgument(gql_field_cls)  # TODO: set default_value
        return arg

    @staticmethod
    def _parameters_for(f):

        def _annotation_type_of(param):
            annotation = param.annotation
            import inspect
            if annotation is inspect._empty:
                import warnings
                warnings.warn(f'parameter {param} is not annotated'
                                 ' for conversion to graphql argument')
                return None
            return annotation

        from inspect import signature
        sig = signature(f)
        param_d = OrderedDict(sig.parameters.items())
        # param_d.popitem(last=False)  # Pop self argument
        # NOTE: param_d.popitem(last=False) may not always pop self argument
        for name, param in param_d.items():
            typ = _annotation_type_of(param)
            if typ is not None:
                yield name, typ
        yield from ()

    def _args_of_f(self, f):
        for name, annotated_type in self._parameters_for(f):
            yield name, self._param_to_graphql_arg(annotated_type=annotated_type)
        yield from ()

    def _args_of(self, rule):
        f = self._f_of_rule(func_name=rule)
        return self._args_of_f(f=f)

    op_type = None

    def _as_graphql_schema(self):
        from onto.models.utils import \
            _graphql_object_type_from_attributed_class
        attributed = self.view_model_cls

        ot = attributed.get_graphql_object_type(is_input=False)

        if self.many:
            ot = graphql.GraphQLList(type_=ot)

        name, args = self._register_op()

        return graph_schema(
            op_type=self.op_type,
            name=name,
            graphql_object_type=ot,
            args=args
        )

    def start(self):
        subscription_schema = self._as_graphql_schema()
        return subscription_schema

    _get_user = staticmethod(get_user)



class GraphQLSubscriptionSink(GraphQLSink):

    op_type = 'Subscription'

    @staticmethod
    def _get_user(info):
        return info.context.user

    async def publish(self, *args, **kwargs):
        return await self.broker.publish(*args, **kwargs)

    def start(self, loop):
        _ = asyncio.run_coroutine_threadsafe(
            self.broker.source_while_loop(), loop=loop
        )
        return super().start()

    def _register_op(self):
        from gql import subscribe
        async def f(parent, info, **kwargs):
            kwargs = {
                'user': self._get_user(info),
                'info': info,
                **kwargs,
            }
            # Register topic
            topic_name = await self._invoke_mediator(func_name='add_topic', **kwargs)

            # Perform before_subscription
            await self._invoke_mediator(func_name='before_subscription', **kwargs)

            from onto.utils import random_id
            sub_id = random_id()

            self.broker.bind(pub_id=topic_name, sub_id=sub_id)
            async for event in self.broker.sink_while_loop(sub_id=sub_id):
                yield {
                    self.sink_name:
                        self._invoke_mediator(func_name='on_event',
                                              event=event, **kwargs)
                }

        # name = self.parent().__name__
        # f.__name__ = name
        name = self.sink_name
        f.__name__ = name
        subscribe(f)
        args = dict(self._args_of('add_topic'))

        return name, args


class GraphQLQuerySink(GraphQLSink):

    op_type = 'Query'

    def _register_op(self):
        from gql import query
        extra_args = set()

        async def f(parent, info, **kwargs):
            kwargs = {
                'info': info,
                **kwargs,
            }

            try:
                user = self._get_user(info)
                kwargs['user'] = user
            except:
                import logging
                logging.error('未能解析user，可能是没有装载 AuthMiddleware；程序将继续执行以兼容不需要用户的测试代码')

            res = await self._invoke_mediator(func_name='query', **kwargs)
            return res

        name = self.sink_name
        f.__name__ = name
        query(f)
        args = dict(self._args_of('query'))
        return name, args


class GraphQLMutationSink(GraphQLSink):

    op_type = 'Mutation'

    def _register_op(self):
        from gql import mutate

        extra_args = set()
        async def f(parent, info, **kwargs):
            kwargs = {
                'user': self._get_user(info),
                'info': info,
                **kwargs
            }
            with graphql_context(info):
                res = await self._invoke_mediator(func_name='mutate', **kwargs)
            return res

        name = self.sink_name
        f.__name__ = name
        mutate(f, snake_argument=False)
        args = dict(self._args_of('mutate'))
        return name, args


def op_schema(op_type, schema_all):
    from onto.helpers.graphql import default_field_resolver

    ot = graphql.GraphQLObjectType(
        name=op_type,
        fields={
            schema.name:
                graphql.GraphQLField(
                    schema.graphql_object_type,
                    resolve=default_field_resolver,
                    args=schema.args)
            for schema in schema_all
            if schema.op_type == op_type
        }
    )
    return ot


query = GraphQLQuerySink
mutation = GraphQLMutationSink
subscription = GraphQLSubscriptionSink



class Protocol:

    def __init__(self):
        self.mapping = dict()

    def _register(self, rule):
        def decorator(f):
            self.mapping[rule] = f.__name__
            return f
        return decorator

    def fname_of(self, rule):
        if rule in self.mapping:
            return self.mapping[rule]
        else:
            return None

    def __getattr__(self, item):
        return self._register(item)



from onto.source.base import Source


class FirestoreSource(Source):

    def __init__(self, query):
        """ Initializes a ViewMediator to declare protocols that
                are called when the results of a query change. Note that
                mediator.start must be called later.

        :param query: a listener will be attached to this query
        """
        super().__init__()
        self.query = query

    def start(self):
        self._register()

    def _register(self):
        from onto.context import Context as CTX
        CTX.db.listener().register(query=self.query, source=self)

    @classmethod
    def delta(cls, container):
        start = container._read_times[-2]
        end = container._read_times[-1]
        for key in container.d.keys():
            for snapshot in container.get_with_range(key, start, end):
                from onto.database import Snapshot
                prev: Snapshot = snapshot.prev
                cur: Snapshot = snapshot
                if not prev.exists:
                    yield ("on_create", key, cur)
                elif prev.exists and cur.exists:
                    yield ("on_update", key, cur)
                elif prev.exists and not cur.exists:
                    yield ("on_delete", key, cur)
                else:
                    raise ValueError

    def _call(self, container):
        with container.lock:
            for func_name, ref, snapshot in self.delta(container):
                self._invoke_mediator(
                    func_name=func_name, ref=ref, snapshot=snapshot)



from typing import List
from onto.database import Reference, Snapshot, Database
from onto.store.snapshot_container import SnapshotContainer
from onto.context import Context as CTX


class Gallery:

    def __init__(self, database=None):
        self.tasks = dict()  # TODO: Watch out for when (doc_ref, obj_type_super) and (doc_ref, obj_type_sub) are both in the set; the objects will be equivalent, but initialized twice under the current plan
        self.visited = set()
        self.container = SnapshotContainer()
        self.object_container = dict()
        self.save_tasks = dict()
        if database is None:
            database = CTX.db
        self._database = database

    def _datastore(self):
        return self._database

    def insert(self, *, doc_ref, obj_type) -> None:
        """
        TODO: implement obj_type kwarg or toss it
        :param doc_ref:
        :param obj_type:
        :return:
        """
        if doc_ref in self.tasks or doc_ref in self.visited:
            return
        self.tasks[doc_ref] = obj_type
        self.visited.add(doc_ref)

    def save_later(self, obj, transaction=None, **kwargs):
        forward_kwargs = dict(kwargs, transaction=transaction)
        if obj.doc_ref not in self.save_tasks:
            self.save_tasks[obj.doc_ref] = (obj, forward_kwargs)
        else:
            # TODO: implement
            # TODO: add logic for checking changes
            pass

    def save(self):
        for _, (obj, kwargs) in self.save_tasks.items():
            d = obj._export_as_dict(**kwargs)
            snapshot = Snapshot(d)
            self._datastore().set(snapshot=snapshot, ref=obj.doc_ref, transaction=kwargs['transaction'])

    @staticmethod
    def _get_snapshots_with_listener(refs: List[Reference]):
        raise NotImplementedError
        # from onto.context import Context as CTX
        #
        # def cb(*args, **kwargs):
        #     print(f"{args} {kwargs}")
        # CTX.listener.from_refs(refs, cb=cb)


    @staticmethod
    def _get_snapshots_with_batch(database: Database, transaction, **kwargs):
        """ needed because transactional wrapper uses specific argument
            ordering

        :param transaction:
        :param kwargs:
        :return:
        """
        return database.get_many(transaction=transaction, **kwargs)

    def refresh(self, transaction, get_snapshots=None):

        if get_snapshots is None:
            get_snapshots = self._get_snapshots_with_batch

        while len(self.tasks) != 0:

            refs = list()
            for doc_ref in self.tasks:
                refs.append(doc_ref)

            res = get_snapshots(database=self._datastore(), refs=refs, transaction=transaction)
            for ref, doc in res:
                self.container.set(key=ref, val=doc)

                # for doc in res:
                obj_type = self.tasks[ref]
                del self.tasks[ref]
                instance = obj_type.from_snapshot(
                    ref=ref, snapshot=doc, _store=self)

                # d = doc.to_dict()
                # obj_cls = resolve_obj_cls(cls=obj_type, d=d)
                #
                # schema_obj = obj_cls.get_schema_obj()
                # d = schema_obj.load(d)
                # d = obj_cls._import_from_dict(d, transaction=transaction,
                #                               _store=self)
                #
                # instance = obj_cls.new(
                #     **d, transaction=transaction)
                self.object_container[ref] = instance

    def retrieve(self, *, doc_ref, obj_type):
        if doc_ref not in self.object_container:
            snapshot = self.container.get(key=doc_ref)
            self.object_container[doc_ref] = obj_type.from_snapshot(ref=doc_ref, snapshot=snapshot)
        return self.object_container[doc_ref]



import time
import warnings

from google.cloud.firestore_v1 import CollectionReference

DEFAULT_WAIT_TIME = 5

import logging
root_log = logging.getLogger()
stats = {"total": 0}

try:
    import os
    wt = os.getenv("WAIT_TIME")
    if wt is not None:
        DEFAULT_WAIT_TIME = int(wt)
except Exception as e:
    """
    TODO: find the best practice to handle a "black swan" error 
    """
    raise e


def _wait(secs=DEFAULT_WAIT_TIME, factor=1):
    secs = secs * factor
    time.sleep(secs)
    stats["total"] += secs
    logging.debug(msg=f"waited for {secs} seconds, "
                f"a total of {stats['total']} seconds in this session")


def _delete_all(CTX, collection_name=None, subcollection_name=None):
    """ DANGEROUS :For testing purposes only. Use with caution.
        Never use in production. Protection against such use case may fail.

    :param CTX:
    :param collection_name:
    :return:
    """
    app_name = CTX.firebase_app.name
    if not app_name.find("testing"):
        raise Exception("Firebase App Name is {}. "
                        "Only app name containing testing is supported"
                        .format(app_name))
    if collection_name is not None:
        collection: CollectionReference = CTX.db.firestore_client.collection(collection_name)
    elif subcollection_name is not None:
        collection: CollectionReference = CTX.db.firestore_client.collection_group(
            subcollection_name)

    warnings.warn("Deleting collection: {}, App Name: {}.".format(collection_name, app_name))

    def delete_collection(coll_ref, batch_size, doc_limit=10):
        """
        Ref: https://firebase.google.com/docs/firestore/manage-data/delete-data
        :param coll_ref:
        :param batch_size:
        :return:
        """
        docs = coll_ref.limit(doc_limit).get()
        deleted = 0
        docs_count = 0

        for doc in docs:
            docs_count = docs_count + 1
            print(u'Deleting doc {} => {}'.format(doc.id, doc.to_dict()))
            doc.reference.delete()
            deleted = deleted + 1

        # Note that this line was "if deleted >= batch_size"
        if deleted < batch_size and docs_count >= doc_limit:
            return delete_collection(coll_ref, batch_size)

    delete_collection(collection, 100)



from collections import namedtuple
from datetime import datetime

from onto.domain_model import DomainModel
from onto.models.base import Serializable
from onto import attrs


class UserTime(Serializable):
    time: datetime = attrs.bproperty()
    timestamp = attrs.integer(import_enabled=False)

    @timestamp.getter
    def timestamp(self):
        return self.time.timestamp()


class Location(DomainModel):

    latitude = attrs.bproperty(type_cls=float)
    longitude = attrs.bproperty(type_cls=float)


class RideOrder(DomainModel):

    category = attrs.string()
    create_time = attrs.embed(obj_cls=UserTime)
    start_location = attrs.relation(dm_cls=Location)


class RideTime(DomainModel):

    order = attrs.relation(dm_cls=RideOrder)
    category = attrs.string(import_enabled=False)
    timestamp = attrs.bproperty(type_cls=float)

    @category.getter
    def category(self):
        return self.order.category

    @timestamp.getter
    def timestamp(self):
        return self.order.time.timestamp


_window = namedtuple('_window', ['lo_excl', 'lo_incl', 'hi_excl', 'hi_incl'])


class Window:

    class now:
        pass

    @classmethod
    def last(cls, *, hour=None):
        return _window(lo_incl=cls.now-hour*3600, high_excl=cls.now)

from pony.orm import avg

class GlobalRideTime(DomainModel):
    """
    We should know that ride time is in poisson distribution
    """

    earlier = Window.now - 3600
    later = Window.now

    @property
    def ride_times(self):
        for rt in RideTime:
            if self.earlier < rt.timestamp <= self.later :
                yield rt

    ts = attrs.bproperty(type_cls=float, collection=list)
    @ts.getter
    def ts(self):
        for rt in self.ride_times:
            yield rt.timestamp - self.earlier

    mean = attrs.bproperty(type_cls=float)
    @mean.getter
    def mean(self):
        return avg(self.ts)

    def poisson_dist(self):
        """

        :return:
            a: lambda value in poisson distribution,
            pvalue: confidence value that the data follows a poisson process
        """
        from scipy.stats import gamma, kstest
        a, floc, scale = gamma.fit(
            self.ts,
            fscale=(self.later-self.earlier),
            floc=self.earlier
        )
        _lambda = scale
        # Note that "Arrival time" has gamma distribution and
        #   "Arrival Interval" has poisson distribution.
        #   We are fitting with gamma dist to reduce computation overhead.
        kstest_res = kstest(self.ts, "gamma", (a, floc, scale))
        pvalue = kstest_res.pvalue
        return a, pvalue

    # lambda_hat = attrs.bproperty(type_cls=float)

    # @lambda_hat.getter
    # def lambda_hat(self):
    #     return self.mean


    n = attrs.integer()
    @n.getter
    def n(self):
        return len(self.ride_times)

    # standard_error = attrs.bproperty(type_cls=float)
    # @standard_error.getter
    # def standard_error(self):
    #     from math import sqrt
    #     return sqrt(self.lambda_hat/self.n)

    chi_sq = attrs.bproperty(type_cls=float)
    @chi_sq.getter
    def chi_sq(self):

        def d(val):
            return (val-self.mean) ** 2

        return avg(d(val) for val in self.ride_times )

